from glob import glob
from datetime import datetime
from collections import namedtuple

import cv2
import torch
import numpy as np
from tifffile import tifffile

from watch.tasks.rutgers_material_change_detection.utils.util_misc import get_crop_slices
from watch.tasks.rutgers_material_change_detection.datasets.base_dataset import BaseDataset


class PassiveEarthObservationDataset(BaseDataset):
    def __init__(
        self,
        root_dir,
        split,
        video_slice,
        task_mode,
        seed_num=0,
        transforms=None,
        normalize_mode=None,
        channels=None,
        max_iterations=None,
    ):
        super().__init__(
            root_dir,
            split,
            video_slice,
            task_mode,
            transforms=transforms,
            seed_num=seed_num,
            normalize_mode=normalize_mode,
            channels=channels,
            max_iterations=max_iterations,
        )

        if task_mode == "ss_arrow_of_time":
            self.ss_aot_reverse_prob = 0.5
            print("WARNING: Need to add arguments for pct reverse this method.")
        elif task_mode == "ss_splice_change":
            self.ss_splice_prob = 0.5
            print("WARNING: Need to add arguments for pct splice method.")

        if (self.normalize_mode == "local") and (task_mode in ["ss_splice_change", "ss_splice_change_index"]):
            print("WARNING: Does not make sense to normalize this task locally.")

        # Figure out how many channels to load.
        Channel_Info = namedtuple("channel_info", ["band_name", "wl_name", "wavelength", "scale_factor"])
        self.channel_info = {}
        if channels == "RGB":
            self.channel_info["B02"] = Channel_Info(band_name="B02", wl_name="blue", wavelength=492.4, scale_factor=1)
            self.channel_info["B03"] = Channel_Info(band_name="B03", wl_name="green", wavelength=559.8, scale_factor=1)
            self.channel_info["B04"] = Channel_Info(band_name="B04", wl_name="red", wavelength=664.6, scale_factor=1)
        elif channels == "RGB_NIR":
            self.channel_info["B02"] = Channel_Info(band_name="B02", wl_name="blue", wavelength=492.4, scale_factor=1)
            self.channel_info["B03"] = Channel_Info(band_name="B03", wl_name="green", wavelength=559.8, scale_factor=1)
            self.channel_info["B04"] = Channel_Info(band_name="B04", wl_name="red", wavelength=664.6, scale_factor=1)
            self.channel_info["B08"] = Channel_Info(band_name="B08", wl_name="nir", wavelength=832.8, scale_factor=1)
        elif channels == "ALL":
            self.channel_info["B01"] = Channel_Info(band_name="B01", wl_name="coastal", wavelength=442, scale_factor=6)
            self.channel_info["B02"] = Channel_Info(band_name="B02", wl_name="blue", wavelength=492, scale_factor=1)
            self.channel_info["B03"] = Channel_Info(band_name="B03", wl_name="green", wavelength=559, scale_factor=1)
            self.channel_info["B04"] = Channel_Info(band_name="B04", wl_name="red", wavelength=664, scale_factor=1)
            self.channel_info["B05"] = Channel_Info(band_name="B05", wl_name="nir", wavelength=704, scale_factor=2)
            self.channel_info["B06"] = Channel_Info(band_name="B06", wl_name="nir", wavelength=740, scale_factor=2)
            self.channel_info["B07"] = Channel_Info(band_name="B07", wl_name="nir", wavelength=783, scale_factor=2)
            self.channel_info["B08"] = Channel_Info(band_name="B08", wl_name="nir", wavelength=832, scale_factor=1)
            self.channel_info["B09"] = Channel_Info(band_name="B09", wl_name="swir", wavelength=945, scale_factor=6)
            self.channel_info["B11"] = Channel_Info(band_name="B11", wl_name="swir", wavelength=1610, scale_factor=2)
            self.channel_info["B12"] = Channel_Info(band_name="B12", wl_name="swir", wavelength=2190, scale_factor=2)
            self.channel_info["B8A"] = Channel_Info(band_name="B8A", wl_name="nir", wavelength=864, scale_factor=2)
        else:
            raise NotImplementedError(f'Channels equal to "{channels}" not implmented.')
        self.n_channels = len(self.channel_info.keys())

        # Get all video directories.
        video_dirs = sorted(glob(root_dir + "/**/crop_*", recursive=True))

        # Compute dataset splits, i.e. train and validation. No test split.
        if split == "test":
            raise ValueError("No test split for PEO dataset, only train and validation.")
        elif split in ["train", "valid"]:
            # Randomly divide video directories with a 90/10 split.
            np.random.shuffle(video_dirs)
            train_split_pct, valid_split_pct = 0.9, 0.1
            assert (train_split_pct + valid_split_pct) == 1.0
            n_videos = len(video_dirs)

            split_num = int(n_videos * train_split_pct)

            if split == "train":
                split_video_dirs = video_dirs[:split_num]
            else:
                split_video_dirs = video_dirs[split_num:]
        else:
            raise NotImplementedError(f'Split name "{split}" is an invalid split type.')

        # Get crop slices.
        # Assume that all band images are size 1098x1098 when resized to highest resolution band.
        og_image_height, og_image_width = 1098, 1098
        crop_slices = get_crop_slices(
            og_image_height, og_image_width, video_slice.height, video_slice.width, step=video_slice.stride
        )

        # Create examples for each video.
        self.dset_slices = []
        self.max_frames = 0
        for video_dir in split_video_dirs:
            # Get all image directories in ascending temporal order.
            image_dirs = glob(video_dir + "/**/")

            ## Get date from image directory names.
            datetimes = []
            for image_dir in image_dirs:
                datetime_str = image_dir.split("/")[-2].split("_")[2]
                year = int(datetime_str[:4])
                month = int(datetime_str[4:6])
                day = int(datetime_str[6:8])
                datetimes.append(datetime(year, month, day))

            ## Sort image directories by datetimes.
            image_dirs = [image_dir for _, image_dir in sorted(zip(datetimes, image_dirs))]
            datetimes = sorted(datetimes)

            # Update the maximum number of frames in a video.
            self.max_frames = max(len(image_dirs), self.max_frames)

            # Get all image paths for each image directory.
            image_dirs_dict = {}
            for image_dir in image_dirs:
                image_paths = sorted(glob(image_dir + "/*.tif"))
                image_dirs_dict[image_dir] = image_paths

            for crop_slice in crop_slices:
                example = {}
                example["image_dirs_dict"] = image_dirs_dict
                example["og_height"] = og_image_height
                example["og_width"] = og_image_width
                example["space_crop_slice"] = crop_slice
                example["datetimes"] = datetimes

                self.dset_slices.append(example)

        # Update max frames based on the number of frames.
        if self.n_frames is None:
            self.n_frames = self.max_frames

        if self.max_frames > self.n_frames:
            self.max_frames = self.n_frames

        # Get number of classes for dataset with given task mode.
        if task_mode == "ss_arrow_of_time":
            self.n_classes = 2
        elif task_mode == "ss_triplet":
            self.n_classes = 0
        elif task_mode == "ss_splice_change":
            self.n_classes = 2
        elif task_mode == "ss_splice_change_index":
            self.n_classes = self.max_frames
        elif task_mode == "ss_mat_recon":
            self.n_classes = self.n_channels
        else:
            raise NotImplementedError(f'Number of classes for this task "{task_mode}" needs to be added.')

    def to_rgb(self, image, gamma=1.0):

        if self.n_channels in [3, 4]:
            # r_band = image[0] * 1.0
            # b_band = image[1] * 1.0
            # g_band = image[2] * 1.0

            r_band = image[0] * 0.85
            b_band = image[1] * 0.85
            g_band = image[2] * 1.45
        elif self.n_channels == 12:
            r_band = image[1]
            b_band = image[2]
            g_band = image[3]
        else:
            raise NotImplementedError(f'Cannot handle "{self.n_channels}" number of channels in to_rgb method.')
        rgb_image = np.stack((r_band, g_band, b_band), axis=2)

        # Adjust color of image.
        # Assuming that image is from the output of __getitem__
        rgb_image *= 10
        rgb_image = np.clip(rgb_image, 0, 1)
        rgb_image = rgb_image ** (1 / gamma)

        rgb_image = np.clip(rgb_image, 0, 1)
        rgb_image *= 255
        rgb_image = rgb_image.astype("uint8")

        return rgb_image

    def get_pixel_normalization_params(self):
        Norm_Params = namedtuple("Norm_Params", ["mean", "std"])
        norm_params_dict = {
            "B01": Norm_Params(mean=-6.7256, std=6.1035),
            "B02": Norm_Params(mean=-6.5524, std=5.9112),  # blue
            "B03": Norm_Params(mean=-6.2216, std=5.5402),  # green
            "B04": Norm_Params(mean=-6.0112, std=5.2998),  # red
            "B05": Norm_Params(mean=-5.6509, std=4.8835),
            "B06": Norm_Params(mean=-5.1046, std=4.2324),
            "B07": Norm_Params(mean=-4.8723, std=3.9443),
            "B08": Norm_Params(mean=-4.7865, std=3.8355),  # nir
            "B09": Norm_Params(mean=-4.5090, std=3.4673),
            "B11": Norm_Params(mean=-4.8664, std=3.9372),
            "B12": Norm_Params(mean=-5.4962, std=4.7033),
            "B8A": Norm_Params(mean=-4.7161, std=3.7464),
        }

        mean_vals, std_vals = [], []
        if self.channels == "RGB":
            for band_name in ["B04", "B03", "B02"]:
                mean_vals.append(norm_params_dict[band_name].mean)
                std_vals.append(norm_params_dict[band_name].std)
        elif self.channels == "RGB_NIR":
            for band_name in ["B04", "B03", "B02", "B08"]:
                mean_vals.append(norm_params_dict[band_name].mean)
                std_vals.append(norm_params_dict[band_name].std)
        elif self.channels == "ALL":
            for band_name in ["B01", "B02", "B03", "B04", "B05", "B06", "B07", "B08", "B09", "B11", "B12", "B8A"]:
                mean_vals.append(norm_params_dict[band_name].mean)
                std_vals.append(norm_params_dict[band_name].std)

        mean_vals, std_vals = np.array(mean_vals), np.array(std_vals)

        # Format shapes.
        mean_vals = mean_vals[np.newaxis, :, np.newaxis, np.newaxis]
        std_vals = std_vals[np.newaxis, :, np.newaxis, np.newaxis]
        return mean_vals, std_vals

    def load_frame(self, channel_image_paths):
        if self.n_channels == 3:
            # Choose RGB bands
            r_channel_img_path = channel_image_paths[3]  # B04, RED
            g_channel_img_path = channel_image_paths[2]  # B03, GREEN
            b_channel_img_path = channel_image_paths[1]  # B02, BLUE

            r_channel_img = tifffile.imread(r_channel_img_path)
            g_channel_img = tifffile.imread(g_channel_img_path)
            b_channel_img = tifffile.imread(b_channel_img_path)

            frame = np.stack([r_channel_img, g_channel_img, b_channel_img], axis=2)
        elif self.n_channels == 4:
            r_channel_img_path = channel_image_paths[3]  # B04, RED
            g_channel_img_path = channel_image_paths[2]  # B03, GREEN
            b_channel_img_path = channel_image_paths[1]  # B02, BLUE
            nir_channel_img_path = channel_image_paths[7]  # B08, NIR

            r_channel_img = tifffile.imread(r_channel_img_path)
            g_channel_img = tifffile.imread(g_channel_img_path)
            b_channel_img = tifffile.imread(b_channel_img_path)
            nir_channel_img = tifffile.imread(nir_channel_img_path)

            frame = np.stack([r_channel_img, g_channel_img, b_channel_img, nir_channel_img], axis=2)

        elif self.n_channels == 12:

            height, width = tifffile.imread(channel_image_paths[3]).shape

            channel_images = []
            for img_path in channel_image_paths:
                channel_image = tifffile.imread(img_path)

                h_img, w_img = channel_image.shape

                if (h_img != height) or (w_img != width):
                    channel_image = cv2.resize(channel_image, dsize=(width, height))

                channel_images.append(channel_image)

            frame = np.stack(channel_images, axis=2)
            assert (frame.shape[0] == height) and (frame.shape[1] == width)

        return frame

    def load_frames(self, image_dirs, crop_slice, og_height, og_width):
        """Create a video given a set of image directories and crop parameters.

        Args:
            image_dirs (list(str)): A list containing the directories for all image channels paths in a region.
            crop_slice (tuple): A tuple containing [top_height, left_width, height_size, width_size]

        Returns:
            frames (list(np.array)): A list of numpy arrays.
            frame_indices (list(int)): a list of frame indices.
        """
        h0, w0, h, w = crop_slice

        frames, frame_indices = [], []
        if self.max_frames == 2:
            # image_dirs = [image_dirs[0], image_dirs[-1]]
            frame_indices = [0, len(image_dirs)]
        elif self.max_frames >= len(image_dirs):
            frame_indices = np.arange(len(image_dirs))
        else:
            # Randomly sample from all frames.
            indices = np.arange(len(image_dirs))
            frame_indices = np.random.choice(indices, size=self.max_frames, replace=False)
            frame_indices = sorted(frame_indices)

        for i, image_dir in enumerate(image_dirs):

            if i in frame_indices:
                channel_image_paths = sorted(glob(image_dir + "/*.tif"))

                frame = self.load_frame(channel_image_paths)

                # TODO: Figure out a better method for matching image shapes.
                if (frame.shape[0] != og_height) or (frame.shape[1] != og_width):
                    frame = cv2.resize(frame, (og_width, og_height))

                crop = frame[h0 : (h0 + h), w0 : (w0 + w), :]  # [height, width, channels]
                frames.append(crop)

        return frames, frame_indices

    def get_self_supervised_triplet_example(self, index):
        example = self.dset_slices[index]

        # Get anchor and positive items.
        ## Get paths to anchor and postive frames.
        image_dirs = list(example["image_dirs_dict"].keys())

        assert len(image_dirs) >= 2

        if len(image_dirs) == 2:
            anchor_index = 0
            positive_index = 0
        else:
            anchor_index = torch.randint(0, len(image_dirs), size=(1,)).item()

            positive_index = anchor_index
            while positive_index == anchor_index:
                positive_index = torch.randint(0, len(image_dirs), size=(1,)).item()

        ## Get crop information.
        h0, w0, h, w = example["space_crop_slice"]
        hE, wE = h0 + h, w0 + w

        ## Load frames.
        anchor_band_paths = sorted(glob(image_dirs[anchor_index] + "/*.tif"))
        anchor_frame = self.load_frame(anchor_band_paths)  # [height, width, n_channels]
        anchor_frame = anchor_frame[h0:hE, w0:wE, :]

        positive_band_paths = sorted(glob(image_dirs[positive_index] + "/*.tif"))
        positive_frame = self.load_frame(positive_band_paths)  # [height, width, n_channels]
        positive_frame = positive_frame[h0:hE, w0:wE, :]

        # Get negative item.
        ## Get index of random negative example.
        neg_ex_index = index
        while neg_ex_index != index:
            neg_ex_index = torch.randint(0, len(image_dirs), size=(1,)).item()
        neg_example = self.dset_slices[neg_ex_index]

        ## Select image directory for neg example.
        neg_image_dirs = list(neg_example["image_dirs_dict"].keys())
        negative_index = torch.randint(0, len(neg_image_dirs), size=(1,)).item()

        ## Load negative frame.
        negative_band_paths = sorted(glob(image_dirs[negative_index] + "/*.tif"))
        negative_frame = self.load_frame(negative_band_paths)  # [height, width, n_channels]
        negative_frame = negative_frame[h0:hE, w0:wE, :]

        # Apply transforms to triplet items.
        ## Stack tripet items into video of shape [frames, channels, height, width].
        triplet_video = np.stack([anchor_frame, positive_frame, negative_frame], axis=0)
        triplet_video = triplet_video.transpose((0, 3, 1, 2))

        ## Remove any NaN values in video.
        nan_indices = np.isnan(triplet_video)
        triplet_video[nan_indices] = 0

        ## Scale from uint16 to [0,1].
        triplet_video = np.clip(triplet_video, 0, 2**15) / 2**15
        triplet_video = triplet_video.astype("float32")

        # Apply transforms to imagery.
        if self.transforms:
            # Need to create custom transform object
            triplet_video, _ = self.transforms(triplet_video, target=None)

        # Normalize triplet items.
        triplet_video = self.normalize(triplet_video)

        # Pad size of triplet examples.
        if (triplet_video.shape[2] < self.video_slice.height) or (triplet_video.shape[3] < self.video_slice.width):
            canvas = (
                np.ones(
                    (triplet_video.shape[0], triplet_video.shape[1], self.video_slice.height, self.video_slice.width),
                    dtype=triplet_video.dtype,
                )
                * self.pad_value
            )
            canvas[:, :, : triplet_video.shape[2], : triplet_video.shape[3]] = triplet_video
            triplet_video = canvas

        # Scale resolution.
        if self.scale != 1 and self.scale is not None:
            triplet_video, _ = self.scale_video_target(triplet_video, None, self.scale)

        # Convert numpy arrays to tensor.
        # triplet_video = np.copy(triplet_video)
        triplet_video = torch.from_numpy(triplet_video)
        anchor_frame = triplet_video[0]  # [channels, height, width]
        positive_frame = triplet_video[1]  # [channels, height, width]
        negative_frame = triplet_video[2]  # [channels, height, width]

        output = {}
        output["anchor"] = anchor_frame
        output["positive"] = positive_frame
        output["negative"] = negative_frame
        output[self.task_mode] = 0  # dummy variable

        return output

    def get_self_supervised_arrow_of_time_example(self, index):
        example = self.dset_slices[index]

        # Load video.
        frames, frame_indices = self.load_frames(
            example["image_dirs_dict"], example["space_crop_slice"], example["og_height"], example["og_width"]
        )
        video = np.stack(frames, axis=0)  # [frames, height, width, channels]
        video = video.transpose((0, 3, 1, 2))

        # Remove any NaN values in video.
        nan_indices = np.isnan(video)
        video[nan_indices] = 0

        # Scale from uint16 to [0,1].
        video = np.clip(video, 0, 2**15) / 2**15
        video = video.astype("float32")

        # Reverse the arrow of time based on probability.
        coin = torch.rand(1).item()
        if coin <= self.ss_aot_reverse_prob:
            video = video[::-1]
            reverse = torch.tensor(1)
        else:
            reverse = torch.tensor(0)

        # Apply transforms to imagery.
        if self.transforms:
            # Need to create custom transform object
            video, _ = self.transforms(video, target=None)

        # Normalize video.
        video = self.normalize(video)

        # Pad video to desired size.
        active_frames = torch.zeros(self.max_frames)
        active_frames[: video.shape[0]] = torch.ones(video.shape[0])

        if (video.shape[2] < self.video_slice.height) or (video.shape[3] < self.video_slice.width):
            canvas = (
                np.ones(
                    (video.shape[0], video.shape[1], self.video_slice.height, self.video_slice.width), dtype=video.dtype
                )
                * self.pad_value
            )
            canvas[:, :, : video.shape[2], : video.shape[3]] = video
            video = canvas

        pad_video = (
            np.ones([self.max_frames, video.shape[1], video.shape[2], video.shape[3]], dtype=video.dtype)
            * self.pad_value
        )
        pad_video[: video.shape[0], ...] = video
        video = pad_video

        # Scale resolution.
        if self.scale != 1 and self.scale is not None:
            video, _ = self.scale_video_target(video, None, self.scale)

        # Convert from numpy array to tensor.
        video = np.copy(video)
        video = torch.from_numpy(video)

        # Create output dictionary.
        output = {}
        output["video"] = video
        output[self.task_mode] = reverse  # Int [0 or 1]
        output["datetimes"] = example["datetimes"]
        output["active_frames"] = active_frames
        return output

    def get_self_supervised_splice_change_example(self, index):
        example = self.dset_slices[index]

        # Load original video.
        frames, _ = self.load_frames(
            example["image_dirs_dict"], example["space_crop_slice"], example["og_height"], example["og_width"]
        )
        video = np.stack(frames, axis=0)  # [frames, height, width, channels]
        video = video.transpose((0, 3, 1, 2))

        ## Remove any NaN values in video.
        nan_indices = np.isnan(video)
        video[nan_indices] = 0

        ## Scale from uint16 to [0,1].
        video = np.clip(video, 0, 2**15) / 2**15
        video = video.astype("float32")

        ## Normalize orignal video.
        video = self.normalize(video)

        ## Pad video to desired resolution.
        video = self.pad_video_height_width(video, self.video_slice.height, self.video_slice.width)

        # Possibly load another video to splice.
        coin = torch.rand(1).item()
        if coin <= self.ss_splice_prob:
            # Load another video.

            ## Update target.
            splice_value = torch.tensor(1)

            ## Get another example index from dataset.
            other_example_index = index
            indices = list(range(0, self.__len__()))
            indices.remove(index)
            other_example_index = np.random.choice(indices, size=1, replace=False)[0]

            example = self.dset_slices[other_example_index]

            ## Load video and process it.
            frames, _ = self.load_frames(
                example["image_dirs_dict"], example["space_crop_slice"], example["og_height"], example["og_width"]
            )
            splice_video = np.stack(frames, axis=0)  # [frames, height, width, channels]
            splice_video = splice_video.transpose((0, 3, 1, 2))  # [frames, channels, height, width]

            ### Scale from uint16 to [0,1].
            splice_video = np.clip(splice_video, 0, 2**15) / 2**15
            splice_video = splice_video.astype("float32")

            ### Normalize orignal video.
            splice_video = self.normalize(splice_video)

            ### Pad video to desired resolution.
            splice_video = self.pad_video_height_width(splice_video, self.video_slice.height, self.video_slice.width)

            # Splice videos together.
            vid_1_size = np.random.randint(1, video.shape[0])
            # vid_2_size = np.random.randint(1, splice_video.shape[0] - 1)
            vid_2_size = self.max_frames - vid_1_size
            vid_1_first_index = np.random.randint(0, video.shape[0] - vid_1_size)
            vid_2_first_index = np.random.randint(0, splice_video.shape[0] - vid_2_size)

            vid_1 = video[vid_1_first_index : vid_1_first_index + vid_1_size]
            vid_2 = splice_video[vid_2_first_index : vid_2_first_index + vid_2_size]

            video = np.concatenate((vid_1, vid_2), axis=0)

        else:
            # Dont load another video to splice.
            ## Randomly sample another video.
            splice_value = torch.tensor(0)

        # Apply transforms to imagery.
        if self.transforms:
            # Need to create custom transform object
            video, _ = self.transforms(video, target=None)

        # Pad video to desired frames.
        active_frames = torch.zeros(self.max_frames)
        active_frames[: video.shape[0]] = torch.ones(video.shape[0])

        pad_video = (
            np.ones([self.max_frames, video.shape[1], video.shape[2], video.shape[3]], dtype=video.dtype)
            * self.pad_value
        )
        pad_video[: video.shape[0], ...] = video
        video = pad_video

        # Scale resolution.
        if self.scale != 1 and self.scale is not None:
            video, _ = self.scale_video_target(video, None, self.scale)

        # Convert from numpy array to tensor.
        video = np.copy(video)
        video = torch.from_numpy(video)

        # Create output dictionary.
        output = {}
        output["video"] = video
        output[self.task_mode] = splice_value  # Int [0 or 1]
        output["datetimes"] = example["datetimes"]
        output["active_frames"] = active_frames
        return output

    def get_self_supervised_splice_change_index_example(self, index):
        example = self.dset_slices[index]

        # Load original video.
        frames, _ = self.load_frames(
            example["image_dirs_dict"], example["space_crop_slice"], example["og_height"], example["og_width"]
        )
        video = np.stack(frames, axis=0)  # [frames, height, width, channels]
        video = video.transpose((0, 3, 1, 2))

        ## Remove any NaN values in video.
        nan_indices = np.isnan(video)
        video[nan_indices] = 0

        ## Scale from uint16 to [0,1].
        video = np.clip(video, 0, 2**15) / 2**15
        video = video.astype("float32")

        ## Normalize orignal video.
        video = self.normalize(video)

        ## Pad video to desired resolution.
        video = self.pad_video_height_width(video, self.video_slice.height, self.video_slice.width)

        # Load another video to splice.

        ## Get another example index from dataset.
        other_example_index = index
        indices = list(range(0, self.__len__()))
        indices.remove(index)
        other_example_index = np.random.choice(indices, size=1, replace=False)[0]

        example = self.dset_slices[other_example_index]

        ## Load video and process it.
        frames, _ = self.load_frames(
            example["image_dirs_dict"], example["space_crop_slice"], example["og_height"], example["og_width"]
        )
        splice_video = np.stack(frames, axis=0)  # [frames, height, width, channels]
        splice_video = splice_video.transpose((0, 3, 1, 2))  # [frames, channels, height, width]

        ### Scale from uint16 to [0,1].
        splice_video = np.clip(splice_video, 0, 2**15) / 2**15
        splice_video = splice_video.astype("float32")

        ### Normalize orignal video.
        splice_video = self.normalize(splice_video)

        ### Pad video to desired resolution.
        splice_video = self.pad_video_height_width(splice_video, self.video_slice.height, self.video_slice.width)

        # Splice videos together.
        vid_1_size = np.random.randint(1, video.shape[0])
        # vid_2_size = np.random.randint(1, splice_video.shape[0] - 1)
        vid_2_size = self.max_frames - vid_1_size
        vid_1_first_index = np.random.randint(0, video.shape[0] - vid_1_size)
        vid_2_first_index = np.random.randint(0, splice_video.shape[0] - vid_2_size)

        vid_1 = video[vid_1_first_index : vid_1_first_index + vid_1_size]
        vid_2 = splice_video[vid_2_first_index : vid_2_first_index + vid_2_size]

        video = np.concatenate((vid_1, vid_2), axis=0)

        # Apply transforms to imagery.
        if self.transforms:
            # Need to create custom transform object
            video, _ = self.transforms(video, target=None)

        # Pad video to desired frames.
        active_frames = torch.zeros(self.max_frames)
        active_frames[: video.shape[0]] = torch.ones(video.shape[0])

        pad_video = (
            np.ones([self.max_frames, video.shape[1], video.shape[2], video.shape[3]], dtype=video.dtype)
            * self.pad_value
        )
        pad_video[: video.shape[0], ...] = video
        video = pad_video

        # Scale resolution.
        if self.scale != 1 and self.scale is not None:
            video, _ = self.scale_video_target(video, None, self.scale)

        # Convert from numpy array to tensor.
        video = np.copy(video)
        video = torch.from_numpy(video)

        # Create output dictionary.
        output = {}
        output["video"] = video
        output[self.task_mode] = torch.tensor(vid_1_size)  # Int [0 or 1]
        output["datetimes"] = example["datetimes"]
        output["active_frames"] = active_frames
        return output

    def get_self_supervised_material_reconstruction_example(self, index):
        example = self.dset_slices[index]

        # Load video.
        frames, frame_indices = self.load_frames(
            example["image_dirs_dict"], example["space_crop_slice"], example["og_height"], example["og_width"]
        )
        video = np.stack(frames, axis=0)  # [frames, height, width, channels]
        video = video.transpose((0, 3, 1, 2))

        # Remove any NaN values in video.
        nan_indices = np.isnan(video)
        video[nan_indices] = 0

        # Scale from uint16 to [0,1].
        video = np.clip(video, 0, 2**15) / 2**15
        video = video.astype("float32")

        # Apply transforms to imagery.
        if self.transforms:
            # Need to create custom transform object
            video, _ = self.transforms(video, target=None)

        # Normalize video.
        video = self.normalize(video)

        # Pad video to desired size.
        active_frames = torch.zeros(self.max_frames)
        active_frames[: video.shape[0]] = torch.ones(video.shape[0])

        if (video.shape[2] < self.video_slice.height) or (video.shape[3] < self.video_slice.width):
            canvas = (
                np.ones(
                    (video.shape[0], video.shape[1], self.video_slice.height, self.video_slice.width), dtype=video.dtype
                )
                * self.pad_value
            )
            canvas[:, :, : video.shape[2], : video.shape[3]] = video
            video = canvas

        pad_video = (
            np.ones([self.max_frames, video.shape[1], video.shape[2], video.shape[3]], dtype=video.dtype)
            * self.pad_value
        )
        pad_video[: video.shape[0], ...] = video
        video = pad_video

        # Scale resolution.
        if self.scale != 1 and self.scale is not None:
            video, _ = self.scale_video_target(video, None, self.scale)

        # Convert from numpy array to tensor.
        video = np.copy(video)
        video = torch.from_numpy(video)

        # Create output dictionary.
        output = {}
        output["video"] = video
        output[self.task_mode] = video
        output["datetimes"] = example["datetimes"]
        output["active_frames"] = active_frames
        return output


if __name__ == "__main__":
    from watch.tasks.rutgers_material_change_detection.utils.util_paths import get_dataset_root_dir
    from watch.tasks.rutgers_material_change_detection.utils.util_misc import generate_video_slice_object

    root_dir = get_dataset_root_dir("peo")

    # Visualize one example of dataset.
    root_dir = get_dataset_root_dir("peo")
    task_mode = "ss_arrow_of_time"
    video_slice = generate_video_slice_object(height=300, width=300, n_frames=None, scale=1)
    dset = PassiveEarthObservationDataset(root_dir, "train", video_slice, task_mode, channels="RGB_NIR")
    ex_index = 0
    ex = dset.__getitem__(ex_index)
    dset.visualize_example(
        ex_index, save_path=f"peo_ex_{str(ex_index).zfill(2)}.png", overlay_last_anno=True, num_plot_images=6
    )
    print("TEST 1: Create a visualizion. | PASSED")

    # Test 2: Get a single item from the training set in ss_triplet mode.
    video_slice = generate_video_slice_object(height=150, width=150, n_frames=None, scale=2)
    split = "train"
    task_mode = "ss_triplet"
    dset = PassiveEarthObservationDataset(root_dir, split, video_slice, task_mode, channels="RGB_NIR")
    example = dset.__getitem__(0)
    print("Test 2: Get one example from train split of dataset. | PASSED")

    # Test 3: Get a single item from the validation set in ss_triplet mode.
    video_slice = generate_video_slice_object(height=300, width=300, n_frames=None, scale=1)
    split = "valid"
    task_mode = "ss_triplet"
    dset = PassiveEarthObservationDataset(root_dir, split, video_slice, task_mode, channels="RGB_NIR")
    example = dset.__getitem__(0)
    print("Test 3: Get one example from valid split of dataset. | PASSED")

    # Test 4: Get a single item from the train set in ss_arrow_of_time mode.
    video_slice = generate_video_slice_object(height=300, width=300, n_frames=None, scale=1)
    split = "train"
    task_mode = "ss_arrow_of_time"
    dset = PassiveEarthObservationDataset(root_dir, split, video_slice, task_mode, channels="RGB_NIR")
    example = dset.__getitem__(0)
    print("Test 4: Get one example from valid split of dataset in arrow_of_time mode. | PASSED")

    # Test 5: Visualize an ss_splice_change_index example.
    video_slice = generate_video_slice_object(height=300, width=300, n_frames=10, scale=1)
    split = "train"
    task_mode = "ss_splice_change_index"
    dset = PassiveEarthObservationDataset(root_dir, split, video_slice, task_mode, channels="RGB_NIR")
    ex_index = 52
    example = dset.__getitem__(ex_index)
    dset.visualize_example(
        ex_index,
        save_path=f"peo_ss_splice_change_index_{str(ex_index).zfill(2)}.png",
        overlay_last_anno=False,
        num_plot_images=video_slice.n_frames,
    )
    print("Test 5: Visualize an ss_splice_change_index example.| PASSED")



"""
"""
import re
import ubelt as ub
import scriptconfig as scfg
import json


class MergeRegionModelConfig(scfg.Config):
    """
    Combine the specific features from multiple region files into a single one.
    """
    __default__ = {
        'src':
        scfg.Value([],
                   help='paths to input geojson region files',
                   nargs='+',
                   position=1),
        'dst':
        scfg.Value(None, help='path to combined multi-region file'),
        'match_type':
        scfg.Value('region',
                   help=ub.paragraph('''
            regex that filters results to only contain features where
            properties.type match.
            ''')),
        'match_subtype':
        scfg.Value('site',
                   help=ub.paragraph('''
            regex that filters results to include subregion features as
            metadata inside regions, where properties.type match.
            ''')),
    }
    epilog = r"""
    Example Usage:

        DVC_DPATH=$HOME/data/dvc-repos/smart_watch_dvc

        python -m watch merge_region_models \
            --src $DVC_DPATH/drop1/region_models/*.geojson \
            --dst $DVC_DPATH/drop1/all_regions.geojson \
            --match_type "region"

        python -m watch.cli.merge_region_models \
            --src $DVC_DPATH/drop1/region_models/*.geojson \
            --dst $DVC_DPATH/drop1/all_regions.geojson
    """


def main(cmdline=False, **kwargs):
    r"""

    CommandLine:
        DVC_DPATH=$HOME/data/dvc-repos/smart_watch_dvc \
            xdoctest -m watch.cli.merge_region_models main

    Example:
        >>> # xdoctest: +REQUIRES(env:DVC_DPATH)
        >>> import os
        >>> import glob
        >>> dvc_repo = os.environ.get('DVC_DPATH')
        >>> region_fpath = os.path.join(dvc_repo, 'drop1/region_models')
        >>> json_paths = list(glob.glob(os.path.join(region_fpath, '*.geojson')))
        >>> kwargs = {'src': json_paths}
        >>> main(**kwargs)

    """
    import geojson
    config = MergeRegionModelConfig(data=kwargs, cmdline=cmdline)
    # print('config = {}'.format(ub.urepr(dict(config), nl=1)))

    json_paths = config['src']
    output_fpath = config['dst']

    combo = combine_region_models(json_paths, config['match_type'],
                                  config['match_subtype'])

    if output_fpath is None:
        print(geojson.dumps(combo, indent='    '))
    else:
        print('write to output_fpath = {!r}'.format(output_fpath))
        with open(output_fpath, 'w') as file:
            geojson.dump(combo, file, indent='    ')


__config__ = MergeRegionModelConfig


def combine_region_models(json_paths, match_type=None, match_subtype=None):
    import geojson
    if match_type is None:
        match_re = None
    else:
        match_re = re.compile(match_type)

    if match_subtype is None:
        match_subre = None
    else:
        match_subre = re.compile(match_subtype)

    # Collect features with the "region" type
    all_region_features = []
    for json_fpath in json_paths:
        # Inside each file, there should be one "region" and several "site"s
        # Save the sites inside the region
        with open(json_fpath, 'r') as file:
            data = json.load(file)
        assert data['type'] == 'FeatureCollection'
        collection = geojson.FeatureCollection(**data)
        region_features = []
        site_features = []
        for feat in collection['features']:
            if match_re is None or match_re.match(feat['properties']['type']):
                region_features.append(feat)
            if match_subre is not None and match_subre.match(
                    feat['properties']['type']):
                site_features.append(feat)
        # Check if we have the expected no. of regions/sites
        if len(region_features) == 1:
            region_feature = region_features[0]
            region_feature['properties']['sites'] = site_features
            all_region_features.append(region_feature)
        else:
            if len(site_features) > 0:
                print(
                    'WARNING: discarding sites belonging to ambiguous region')
            all_region_features.extend(region_features)

    combo = geojson.FeatureCollection(features=all_region_features)
    return combo


if __name__ == '__main__':
    """
    """
    main(cmdline=True)



import os
import ubelt as ub

import cv2
import torch
import kwcoco
import kwimage
import numpy as np
from datetime import datetime
from collections import namedtuple
from matplotlib.colors import to_rgba

from watch.tasks.rutgers_material_change_detection.utils.util_misc import get_crop_slices
from watch.tasks.rutgers_material_change_detection.datasets.base_dataset import BaseDataset


class IARPA_SC_EVAL_DATASET(BaseDataset):
    def __init__(
        self,
        kwcoco_path,
        split,
        video_slice,
        task_mode,
        seed_num=0,
        sensor_type="S2",
        channels=None,
        transforms=None,
        normalize_mode=None,
        max_iterations=None,
    ):
        """Constructor.

        Args:
            kwcoco_path (str):
                File path to kwcoco file.

            video_slice (namedtuple):

            task_mode (str, str):
                Name of labels to produce for this dataset. Defaults to 'total_bin_change'.

            sensor_type (str, optional):
                The name of sensor to return image data from. Defaults to 'S2'.
        """

        kwcoco_path = ub.Path(kwcoco_path)
        super().__init__(
            os.fspath(kwcoco_path.parent),
            split,
            video_slice,
            task_mode,
            transforms=transforms,
            seed_num=seed_num,
            normalize_mode=normalize_mode,
            channels=channels,
            max_iterations=max_iterations,
        )

        if split != "valid":
            raise NotImplementedError("Dataset designed for split other than valid.")

        # Figure out how many channels to load.
        Channel_Info = namedtuple("channel_info", ["band_name", "wl_name", "wavelength", "scale_factor"])
        self.channel_info = {}
        if channels == "RGB":
            self.channel_info["B02"] = Channel_Info(band_name="B02", wl_name="blue", wavelength=492.4, scale_factor=1)
            self.channel_info["B03"] = Channel_Info(band_name="B03", wl_name="green", wavelength=559.8, scale_factor=1)
            self.channel_info["B04"] = Channel_Info(band_name="B04", wl_name="red", wavelength=664.6, scale_factor=1)
        elif channels == "RGB_NIR":
            self.channel_info["B02"] = Channel_Info(band_name="B02", wl_name="blue", wavelength=492.4, scale_factor=1)
            self.channel_info["B03"] = Channel_Info(band_name="B03", wl_name="green", wavelength=559.8, scale_factor=1)
            self.channel_info["B04"] = Channel_Info(band_name="B04", wl_name="red", wavelength=664.6, scale_factor=1)
            self.channel_info["B08"] = Channel_Info(band_name="B08", wl_name="nir", wavelength=832.8, scale_factor=1)
        elif channels == "ALL":
            self.channel_info["B01"] = Channel_Info(band_name="B01", wl_name="coastal", wavelength=442, scale_factor=6)
            self.channel_info["B02"] = Channel_Info(band_name="B02", wl_name="blue", wavelength=492, scale_factor=1)
            self.channel_info["B03"] = Channel_Info(band_name="B03", wl_name="green", wavelength=559, scale_factor=1)
            self.channel_info["B04"] = Channel_Info(band_name="B04", wl_name="red", wavelength=664, scale_factor=1)
            self.channel_info["B05"] = Channel_Info(band_name="B05", wl_name="nir", wavelength=704, scale_factor=2)
            self.channel_info["B06"] = Channel_Info(band_name="B06", wl_name="nir", wavelength=740, scale_factor=2)
            self.channel_info["B07"] = Channel_Info(band_name="B07", wl_name="nir", wavelength=783, scale_factor=2)
            self.channel_info["B08"] = Channel_Info(band_name="B08", wl_name="nir", wavelength=832, scale_factor=1)
            self.channel_info["B09"] = Channel_Info(band_name="B09", wl_name="swir", wavelength=945, scale_factor=6)
            self.channel_info["B10"] = Channel_Info(band_name="B10", wl_name="swir", wavelength=1375, scale_factor=6)
            self.channel_info["B11"] = Channel_Info(band_name="B11", wl_name="swir", wavelength=1610, scale_factor=2)
            self.channel_info["B12"] = Channel_Info(band_name="B12", wl_name="swir", wavelength=2190, scale_factor=2)
            self.channel_info["B8A"] = Channel_Info(band_name="B8A", wl_name="nir", wavelength=864, scale_factor=2)
        else:
            raise NotImplementedError(f'Channels equal to "{channels}" not implmented.')
        self.n_channels = len(self.channel_info.keys())

        # if ("Drop2-Aligned-TA1-2022-02-15" in self.dset_dir) or ("Aligned-Drop3-TA1-2022-03-10" in self.dset_dir):
        #     kwcoco_path = os.path.join(root_dir, "data_vali.kwcoco.json")
        # else:
        #     kwcoco_path = os.path.join(root_dir, "vali_data.kwcoco.json")

        # Load kwcoco dataset
        self.coco_dset = kwcoco.CocoDataset(kwcoco_path)

        # Get number of classes for dataset with given task mode.
        if task_mode == "total_bin_change":
            self.n_classes = 2
        elif task_mode == "ss_mat_recon":
            self.n_classes = self.n_channels
        elif task_mode == "total_sem_change":
            self.n_classes = 4
        elif task_mode == "refine_sc":
            self.n_classes = 4
        else:
            raise NotImplementedError(f'Number of classes for this task "{task_mode}" needs to be added.')

        # Get all video ids
        video_ids = [video_info["id"] for video_info in self.coco_dset.videos().objs]

        # Only get videos and images with desired sensor
        self.video_dataset = {}  # video_id -> image_ids (filtered)
        self.max_frames = 0
        self.examples = []
        for video_id in video_ids:
            # Get image ids for this video
            image_ids = self.coco_dset.index.vidid_to_gids[video_id]

            # Sort image_ids by datetime.
            datetimes = [self.get_datetime(image_id) for image_id in image_ids]
            image_ids = [img_id for dt, img_id in sorted(zip(datetimes, image_ids))]

            # Filter based on desired sensor type
            sesnor_image_ids = []
            for image_id in image_ids:
                img_sensor_type = self.coco_dset.index.imgs[image_id]["sensor_coarse"]
                if img_sensor_type == sensor_type:
                    sesnor_image_ids.append(image_id)

            # Update maximum number of frames.
            self.max_frames = max(len(sesnor_image_ids), self.max_frames)

            # Add to video dataset
            if len(sesnor_image_ids) > 1:
                self.video_dataset[video_id] = sesnor_image_ids

            # Get spatial slices.
            H = self.coco_dset.index.imgs[image_ids[0]]["height"]
            W = self.coco_dset.index.imgs[image_ids[0]]["width"]
            crop_slices = get_crop_slices(
                H, W, self.video_slice.height, self.video_slice.width, step=self.video_slice.stride
            )

            # Get temporal slices.
            temporal_slices, video_indices = self.compute_equal_temporal_sampling(image_ids, video_slice.n_frames)

            # TEST: Make sure that all image ids are covered.
            sampled_image_ids = []
            map(sampled_image_ids.extend, temporal_slices)
            sampled_image_ids = np.unique(sum(temporal_slices, []))

            assert len(image_ids) == len(sampled_image_ids)

            # Get region ID.
            region_name = self.coco_dset.index.videos[video_id]["name"]

            # Poplate slices with more information.
            for crop_slice in crop_slices:
                for temporal_slice, rel_video_indices in zip(temporal_slices, video_indices):
                    crop_info = {
                        "video_id": video_id,
                        "og_height": H,
                        "og_width": W,
                        "space_crop_slice": crop_slice,
                        "region_name": region_name,
                        "image_ids": temporal_slice,
                        "relative_video_indices": rel_video_indices,
                    }
                    self.examples.append(crop_info)

        # Update max frames based on the number of frames.
        if self.n_frames is None:
            self.n_frames = self.max_frames

        if self.max_frames > self.n_frames:
            self.max_frames = self.n_frames

    def cid_to_name(self, cid):
        return self.coco_dset.cats[cid]["name"]

    def cid_to_rgb(self, cid):
        return to_rgba(self.coco_dset.cats[cid]["color"])[:3]

    def get_datetime(self, image_id):
        date_captured = self.coco_dset.index.imgs[image_id]["date_captured"]  # '2017-03-23T09:30:31'
        datetime_object = datetime.strptime(date_captured[:10], "%Y-%m-%d")
        return datetime_object

    def compute_equal_temporal_sampling(self, image_ids, frames_per_video):
        # Assuming image IDs are sorted by dates early to latest.
        # abc = self.compute_equal_temporal_sampling([0,1,2,3,4,5,6,7,8,9], 5)
        N = len(image_ids)
        M = frames_per_video

        all_image_ids, temporal_indices = [], []
        for i in range(N):
            indices = list(np.clip(np.arange(i, M + i) - M + 1, 0, N - 1))
            sub_image_ids = list(np.take(image_ids, indices))
            all_image_ids.append(sub_image_ids)
            temporal_indices.append(indices)

        return all_image_ids, temporal_indices

        # if N == M:
        #     return [image_ids]

        # stride = int(np.ceil((N - 2) / (M - 2)))
        # n_samples = int(np.ceil((N - 2) / (M - 2)))

        # first_frame_id, last_frame_id = image_ids[0], image_ids[-1]

        # middle_ids = image_ids[1:-1]
        # # middle_indices = list(range(image_ids[1:-1]))

        # temporal_slices, all_temporal_indices = [], []
        # for index_offset in range(n_samples):
        #     temporal_indices = []
        #     for m in range(M - 2):
        #         sample_index = index_offset + stride * m
        #         sample_index = np.clip(sample_index, 0, N - 2 - 1)
        #         temporal_indices.append(sample_index)

        #     # Get image ids from indices.
        #     temporal_slice = list(np.take(middle_ids, temporal_indices))

        #     # Attach first and last image IDs.
        #     temporal_slice.insert(0, first_frame_id)
        #     temporal_slice.append(last_frame_id)
        #     temporal_slices.append(temporal_slice)

        #     # Get temporal indices.
        #     for i, _ in enumerate(temporal_indices):
        #         temporal_indices[i] += 1

        #     temporal_indices.insert(0, 0)
        #     temporal_indices.append(len(image_ids) - 1)
        #     all_temporal_indices.append(temporal_indices)

        # return temporal_slices, all_temporal_indices

    def get_total_semantic_change_example(self, index):
        crop_info = self.examples[index]

        # Get all image ids for this video
        image_ids = crop_info["image_ids"]
        frames = []
        for image_id in image_ids:
            frame = self.load_frame(image_id, crop_info["space_crop_slice"])
            frames.append(frame)

        ## Stack images
        video = np.stack(frames, axis=0)  # shape: [frames, channels, height, width]

        ## Format image values (remove nans, scale to 0-1)
        # Remove NaNs in images
        nan_indices = np.isnan(video)
        video[nan_indices] = 0

        # Scale from uint16 to [0,1].
        video = np.clip(video, 0, 2**15) / 2**15

        # Compute target label mask.
        target = self.compute_sc_target_mask(
            image_ids, crop_info["og_height"], crop_info["og_width"], crop_info["space_crop_slice"]
        )
        target = np.stack(target, axis=0)

        # Apply transforms to imagery.
        if self.transforms:
            # Need to create custom transform object
            video, target = self.transforms(video, target)

        # Normalize video.
        video, mean, std = self.normalize(video)

        # Pad video to desired window size.
        if (video.shape[2] < self.video_slice.height) or (video.shape[3] < self.video_slice.width):
            canvas = (
                np.ones(
                    (video.shape[0], video.shape[1], self.video_slice.height, self.video_slice.width), dtype=video.dtype
                )
                * self.ignore_index
            )
            canvas[:, :, : video.shape[2], : video.shape[3]] = video
            video = canvas

        # Pad target mask to desired window size.
        if (target.shape[-2] < self.video_slice.height) or (target.shape[-1] < self.video_slice.width):
            canvas = (
                np.ones((target.shape[0], self.video_slice.height, self.video_slice.width), dtype=target.dtype)
                * self.ignore_index
            )
            canvas[:, : target.shape[-2], : target.shape[-1]] = target
            target = canvas

        # Scale resolution of video and target.
        if self.scale != 1 and self.scale is not None:
            video = self.scale_video(video, self.scale, inter_mode=cv2.INTER_NEAREST)
            target = self.scale_frame(target, self.scale)

        # Convert from numpy array to tensor.
        video, target = np.copy(video), np.copy(target)
        video = torch.as_tensor(video)
        target = torch.as_tensor(target)

        ## Pad the number of frames to maximum number of frames.
        # [frames, channels, height, width]
        active_frames = torch.zeros(self.max_frames)
        active_frames[: video.shape[0]] = torch.ones(video.shape[0])

        pad_video = (
            torch.ones([self.max_frames, video.shape[1], video.shape[2], video.shape[3]], dtype=video.dtype)
            * self.ignore_index
        )
        pad_video[: video.shape[0], ...] = video
        video = pad_video

        ## Get metadata from all images in video.
        datetimes = []
        for image_id in image_ids:
            datetimes.append(self.get_datetime(image_id))

        output = {}
        output["video"] = video.float()
        output[self.task_mode] = target
        output["datetimes"] = datetimes
        output["active_frames"] = active_frames
        output["region_name"] = crop_info["region_name"]
        output["crop_info"] = crop_info
        output["mean"] = mean
        output["std"] = std
        return output

    def compute_sc_target_mask(self, image_ids, height, width, spatial_crop):
        h0, w0, h, w = spatial_crop

        anno_images = []
        for image_id in image_ids:
            anno_ids = self.coco_dset.index.gid_to_aids[image_id]
            dets = kwimage.Detections.from_coco_annots(self.coco_dset.annots(anno_ids).objs, dset=self.coco_dset)
            warp_info = self.coco_dset.imgs[image_id]["warp_img_to_vid"]
            if "offset" in warp_info.keys():
                dets = dets.translate(warp_info["offset"])

            anno_canvas = np.zeros([height, width], dtype=np.int32)
            ignore_canvas = np.zeros([height, width], dtype=np.int32)
            ann_polys = dets.data["segmentations"].to_polygon_list()
            ann_cids = [dets.classes.idx_to_id[cidx] for cidx in dets.data["class_idxs"]]

            for polygon, cat_id in zip(ann_polys, ann_cids):
                cid_name = self.cid_to_name(cat_id).lower()
                if cid_name in ["ignore", "unknown", "negative", "positive"]:
                    polygon.fill(ignore_canvas, value=-1)
                elif cid_name in ["no activity", "background"]:
                    pass
                else:
                    polygon.fill(anno_canvas, value=cat_id)
            anno_images.append(anno_canvas[h0 : (h0 + h), w0 : (w0 + w)])
        return anno_images

    def load_frame(self, image_id, crop_slice):
        """Given an image ID and crop parameters, return a single multispectral image.

        Args:
            image_id (int): A integer representing a unique image ID.
            crop_slice (list(int)): A list of four integers corresponding to
                [row_start, col_start, row_end, col_end].

        Returns:
            np.array: A uint16 numpy array of shape [n_channels, height, width] corresponding to a loaded and
            cropped image.
        """
        h0, w0, h, w = crop_slice

        # Get channels.
        if self.channels == "RGB":
            channel_names = "red|blue|green"
        elif self.channels == "RGB_NIR":
            channel_names = "red|blue|green|nir"
        elif self.channels == "ALL":
            channel_names = None

        try:
            delayed_image = self.coco_dset.delayed_load(image_id, channels=channel_names, space="video")
            delayed_crop = delayed_image.crop((slice(h0, h0 + h), slice(w0, w0 + w)))
            frame = delayed_crop.finalize(as_xarray=False).transpose((2, 0, 1))  # [n_channels, height, width]
        except ValueError:
            frame = np.zeros([self.n_channels, h, w], dtype="float")

        return frame

    def get_pixel_normalization_params(self):
        Norm_Params = namedtuple("Norm_Params", ["mean", "std"])
        if "drop2" in self.dset_dir:
            norm_params_dict = {
                "B02": Norm_Params(mean=2.5880, std=2.5634),  # blue
                "B03": Norm_Params(mean=2.3949, std=2.3728),  # green
                "B04": Norm_Params(mean=2.5746, std=2.5507),  # red
                "B08": Norm_Params(mean=4.0442, std=4.0061),  # nir
            }
        elif "TA1" in self.dset_dir:
            norm_params_dict = {
                "B02": Norm_Params(mean=1.5640, std=1.5374),  # blue
                "B03": Norm_Params(mean=1.4271, std=1.4034),  # green
                "B04": Norm_Params(mean=1.5551, std=1.5293),  # red
                "B08": Norm_Params(mean=2.4479, std=2.4086),  # nir
            }
        else:
            norm_params_dict = {
                "B02": Norm_Params(mean=1.7603, std=1.7328),  # blue
                "B03": Norm_Params(mean=2.0488, std=2.0234),  # green
                "B04": Norm_Params(mean=1.9127, std=1.8875),  # red
                "B08": Norm_Params(mean=3.3785, std=3.3383),  # nir
            }
        mean_vals, std_vals = [], []
        if self.channels == "RGB":
            for band_name in ["B04", "B03", "B02"]:
                mean_vals.append(norm_params_dict[band_name].mean)
                std_vals.append(norm_params_dict[band_name].std)
        elif self.channels == "RGB_NIR":
            for band_name in ["B04", "B03", "B02", "B08"]:
                mean_vals.append(norm_params_dict[band_name].mean)
                std_vals.append(norm_params_dict[band_name].std)
        elif self.channels == "ALL":
            raise NotImplementedError("Global normalization parameters not generated for ALL channels.")
        else:
            raise NotImplementedError(f"No global normalization process for channels name {self.channels}")
        mean_vals, std_vals = np.array(mean_vals), np.array(std_vals)

        # Format shapes.
        mean_vals = mean_vals[np.newaxis, :, np.newaxis, np.newaxis]
        std_vals = std_vals[np.newaxis, :, np.newaxis, np.newaxis]
        return mean_vals, std_vals


if __name__ == "__main__":
    # Get root directory of dataset on local machine.
    from watch.tasks.rutgers_material_change_detection.utils.util_paths import get_dataset_root_dir
    from watch.tasks.rutgers_material_change_detection.utils.util_misc import generate_video_slice_object

    dset_name = "iarpa_drop2v2"
    root_dir = get_dataset_root_dir(dset_name)

    # Get dataset split and task_mode.
    split = "valid"
    task_mode = "total_sem_change"

    # Test loading first and last frames of training split.
    video_slice = generate_video_slice_object(height=300, width=300, n_frames=5, scale=1)
    dset = IARPA_SC_EVAL_DATASET(root_dir, split, video_slice, task_mode, channels="RGB_NIR", normalize_mode=None)
    example = dset.__getitem__(0)



"""
mkinit -m watch.tasks.fusion --lazy --noattrs
-w
mkinit -m watch.tasks.fusion --noattrs -w
"""

# Hack to supress pytorch-lightning warning
import setuptools  # NOQA

from watch.tasks.fusion import utils
from watch.tasks.fusion import datamodules
from watch.tasks.fusion import architectures
from watch.tasks.fusion import methods


# Not including these removes the runtime warning:
# 'watch.tasks.fusion.fit' found in sys.modules after import of package
# 'watch.tasks.fusion', but prior to execution of 'watch.tasks.fusion.fit';
# this may result in unpredictable behaviour

# from watch.tasks.fusion import fit
# from watch.tasks.fusion import predict
# from watch.tasks.fusion import evaluate

# We can include them via lazy imports. The other packages are not given lazy
# imports so torch.package can introspect them.


def lazy_import(module_name, submodules, submod_attrs):
    import importlib
    import os
    name_to_submod = {
        func: mod for mod, funcs in submod_attrs.items()
        for func in funcs
    }

    def __getattr__(name):
        if name in submodules:
            attr = importlib.import_module(
                '{module_name}.{name}'.format(
                    module_name=module_name, name=name)
            )
        elif name in name_to_submod:
            submodname = name_to_submod[name]
            module = importlib.import_module(
                '{module_name}.{submodname}'.format(
                    module_name=module_name, submodname=submodname)
            )
            attr = getattr(module, name)
        else:
            raise AttributeError(
                'No {module_name} attribute {name}'.format(
                    module_name=module_name, name=name))
        globals()[name] = attr
        return attr

    if os.environ.get('EAGER_IMPORT', ''):
        for name in name_to_submod.values():
            __getattr__(name)

        for attrs in submod_attrs.values():
            for attr in attrs:
                __getattr__(attr)
    return __getattr__


__getattr__ = lazy_import(
    __name__,
    submodules={
        'architectures',
        'datamodules',
        'evaluate',
        'fit',
        'methods',
        'organize',
        'postprocess',
        'predict',
        'repackage',
        'utils',
    },
    submod_attrs={},
)


def __dir__():
    return __all__


__all__ = ['architectures', 'datamodules', 'evaluate', 'fit', 'methods',
           'organize', 'postprocess', 'predict', 'repackage', 'utils']



"""

Current encoder config names are:

    'smt_it_joint_p8', 'smt_it_joint_n12', 'smt_it_joint_t12', 'smt_it_joint_t24',
    'smt_it_joint_s12', 'smt_it_joint_s24', 'smt_it_joint_m24', 'smt_it_joint_l24',

    'smt_it_stm_p8', 'smt_it_stm_n12', 'smt_it_stm_t12', 'smt_it_stm_t24',
    'smt_it_stm_s12', 'smt_it_stm_s24', 'smt_it_stm_m24', 'smt_it_stm_l24',

    'smt_it_sm_p8', 'smt_it_sm_n12', 'smt_it_sm_t12', 'smt_it_sm_t24',
    'smt_it_sm_s12', 'smt_it_sm_s24', 'smt_it_sm_m24', 'smt_it_sm_l24',

    'smt_it_st_p8', 'smt_it_st_n12', 'smt_it_st_t12', 'smt_it_st_t24',
    'smt_it_st_s12', 'smt_it_st_s24', 'smt_it_st_m24', 'smt_it_st_l24',

    'smt_it_tm_p8', 'smt_it_tm_n12', 'smt_it_tm_t12', 'smt_it_tm_t24',
    'smt_it_tm_s12', 'smt_it_tm_s24', 'smt_it_tm_m24', 'smt_it_tm_l24',

    'smt_it_s_p8', 'smt_it_s_n12', 'smt_it_s_t12', 'smt_it_s_t24',
    'smt_it_s_s12', 'smt_it_s_s24', 'smt_it_s_m24', 'smt_it_s_l24',

    'smt_it_t_p8', 'smt_it_t_n12', 'smt_it_t_t12', 'smt_it_t_t24',
    'smt_it_t_s12', 'smt_it_t_s24', 'smt_it_t_m24', 'smt_it_t_l24',

    'smt_it_hwtm_p8', 'smt_it_hwtm_n12', 'smt_it_hwtm_t12', 'smt_it_hwtm_t24',
    'smt_it_hwtm_s12', 'smt_it_hwtm_s24', 'smt_it_hwtm_m24', 'smt_it_hwtm_l24',

    'smt_it_m_p8', 'smt_it_m_n12', 'smt_it_m_t12', 'smt_it_m_t24',
    'smt_it_m_s12', 'smt_it_m_s24', 'smt_it_m_m24', 'smt_it_m_l24',

    'sm_it_joint_p8', 'sm_it_joint_n12', 'sm_it_joint_t12', 'sm_it_joint_t24',
    'sm_it_joint_s12', 'sm_it_joint_s24', 'sm_it_joint_m24', 'sm_it_joint_l24',

    'sm_it_sm_p8', 'sm_it_sm_n12', 'sm_it_sm_t12', 'sm_it_sm_t24',
    'sm_it_sm_s12', 'sm_it_sm_s24', 'sm_it_sm_m24', 'sm_it_sm_l24'

Notes:
    pip install reformer_pytorch
    pip install performer-pytorch  <- this one
"""

from functools import wraps

import torch
from torch import nn, einsum
import torch.nn.functional as F

import einops
from einops import rearrange, repeat

import ubelt as ub  # NOQA
import math

try:
    import xdev
    profile = xdev.profile
except Exception:
    profile = ub.identity


class ResidualSequential(nn.Sequential):
    """
    A Sequential layer with a residual operation at the end
    """

    def __init__(self, *args):
        super().__init__(*args)

    def forward(self, x):
        return x + super().forward(x)


class ResidualAttentionSequential(ResidualSequential):
    """
    Special case of ResidualSequential to support masking
    """

    def __init__(self, norm, attention):
        super().__init__(norm, attention)

    def forward(self, x, key_padding_mask=None):
        h = x
        h = self[0](h)
        h = self[1](h, key_padding_mask=key_padding_mask)
        return x + h


class MultiheadSelfAttention(torch.nn.MultiheadAttention):
    """
    Inherits from :class:`torch.nn.MultiheadAttention`

    Args:
        embed_dim: total dimension of the model.
        num_heads: parallel attention heads.
        dropout: a Dropout layer on attn_output_weights. Default: 0.0.
        bias: add bias as module parameter. Default: True.
        add_bias_kv: add bias to the key and value sequences at dim=0.
        add_zero_attn: add a new batch of zeros to the key and
                       value sequences at dim=1.
        kdim: total number of features in key. Default: None.
        vdim: total number of features in value. Default: None.
        batch_first: If ``True``, then the input and output tensors are provided
            as (batch, seq, feature). Default: ``False`` (seq, batch, feature).

    Example:
        >>> from watch.tasks.fusion.architectures.transformer import *  # NOQA
        >>> self = MultiheadSelfAttention(4, 1).eval()
        >>> S, B, F = (7, 3, 4)
        >>> x  = (torch.rand(S, B, F) * 10).round()
        >>> # Results should be independent of the batch dim
        >>> y  = self.forward(x)
        >>> y0 = self.forward(x[:, 0:1, :])
        >>> y1 = self.forward(x[:, 1:2, :])
        >>> y2 = self.forward(x[:, 2:3, :])
        >>> assert torch.allclose(y[:, 0:1, :], y0)
        >>> assert torch.allclose(y[:, 1:2, :], y1)
        >>> assert torch.allclose(y[:, 2:3, :], y2)

        >>> key_padding_mask = torch.rand(B, S) > 0.5
        >>> masked_result = self.forward(x, key_padding_mask=key_padding_mask)
    """

    def __init__(self, embed_dim, num_heads, *args, **kwargs):
        super().__init__(embed_dim, num_heads, *args, **kwargs)

    def forward(self, x, key_padding_mask=None):
        """
        Args:
            x (Tensor) : of shape (seq, batch, feature)

            key_padding_mask (Tensor) : of shape (batch, seq).
                A value of True means we will **ignore** the token.

        Returns:
            attn_out : of shape (seq, batch, feature)
        """
        # attention returns a tuple of output and weights, so just take the
        # output
        outs = super().forward(
            query=x, key=x, value=x, key_padding_mask=key_padding_mask)
        attn_out, attn_weights = outs
        return attn_out


try:
    from performer_pytorch import FastAttention

    class FastMultiheadSelfAttention(FastAttention):
        """
        This seems like a good idea, but either I'm using it wrong or the
        C-bindings in normal attention make this lose all of its benefit.

        Ignore:
            D = 9  # embedding dimension
            H = 3   # number of heads
            B = 5   # batch size
            S = 7   # sequence length
            x = torch.rand(S, B, D)
            MultiheadSelfAttention(D, H)(x).shape
            FastMultiheadSelfAttention(D, H)(x)
            from performer_pytorch import FastAttention
            q = einops.rearrange(x, 's b (h e) -> b h s e', h=H)
            FastAttention(dim_heads=D // H, nb_features=None)(q, q, q).shape
        """

        def __init__(self, embed_dim, num_heads):
            self.embed_dim = embed_dim
            self.num_heads = num_heads
            assert embed_dim % num_heads == 0
            dim_heads = embed_dim // num_heads
            nb_features = int(dim_heads * math.log(dim_heads))
            # nb_features = int(dim_heads * 2)
            super().__init__(
                dim_heads, nb_features=nb_features, ortho_scaling=0,
                causal=False, generalized_attention=False, kernel_fn=nn.ReLU(),
                no_projection=False)

        @profile
        def forward(self, x, key_padding_mask=None):
            # import xdev
            # xdev.embed()
            # make compatible with nn.MultiheadAttention
            # s, b, he = x.shape
            # e = self.dim_heads
            # h = self.num_heads
            # Much faster than einops
            if key_padding_mask is not None:
                raise NotImplementedError
            # q = x.contiguous().view(s, b, h, e).permute(1, 2, 0, 3)
            q = einops.rearrange(x, 's b (h e) -> b h s e', e=self.dim_heads)
            # a = FastAttention.forward(self, q, q, q)
            a = super().forward(q, q, q)
            # out = a.permute(2, 1, 0, 3).contiguous().view(s, b, he)
            out = einops.rearrange(a, 'b h s e -> s b (h e)', e=self.dim_heads)
            return out
except ImportError:
    pass


try:
    from reformer_pytorch import LSHSelfAttention

    class ReformerMultiheadedSelfAttention(LSHSelfAttention):
        """
        This seems like a good idea, but either I'm using it wrong or the
        C-bindings in normal attention make this lose all of its benefit.

        Ignore:
            from watch.tasks.fusion.architectures.transformer import *  # NOQA
            D = 9  # embedding dimension
            H = 3   # number of heads
            B = 5   # batch size
            S = 7   # sequence length
            x = torch.rand(S, B, D)

            self = ReformerMultiheadedSelfAttention(D, H)

            MultiheadSelfAttention(D, H)(x).shape
            ReformerMultiheadedSelfAttention(D, H)(x)
            from reformer_pytorch import LSHAttention
            q = einops.rearrange(x, 's b (h e) -> b h s e', h=H)
            FastAttention(dim_heads=D // H, nb_features=None)(q, q, q).shape
        """

        def __init__(self, embed_dim, num_heads):
            self.embed_dim = embed_dim
            self.num_heads = num_heads
            assert embed_dim % num_heads == 0
            dim_heads = embed_dim // num_heads
            self.dim_heads = dim_heads
            # nb_features = int(dim_heads * math.log(dim_heads))
            # nb_features = int(dim_heads * 2)
            super().__init__(
                dim=embed_dim, heads=num_heads, dim_head=dim_heads,
                bucket_size=64, n_hashes=8, causal=False)

        @profile
        def forward(self, x, key_padding_mask=None):
            if key_padding_mask is not None:
                raise NotImplementedError
            s, b, he = x.shape
            bsd = x.permute(1, 0, 2)
            # a = LSHSelfAttention.forward(self, bsd)
            a = super().forward(bsd)
            out = a.permute(1, 0, 2)
            return out
except ImportError:
    pass


def new_attention_layer(embedding_size, n_heads, attention_impl='exact'):
    """
    Example:
        >>> from watch.tasks.fusion.architectures.transformer import *  # NOQA
        >>> import torch
        >>> batch_size = 1
        >>> embedding_size = 4
        >>> n_heads = 2
        >>> num_tokens = 3
        >>> input_shape = (num_tokens, batch_size, embedding_size)
        >>> inputs = torch.rand(*input_shape)
        >>> layer1 = new_attention_layer(embedding_size, n_heads, 'exact')
        >>> outputs1 = layer1(inputs)
        >>> assert outputs1.shape == input_shape
        >>> # xdoctest: +REQUIRES(module:performer_pytorch)
        >>> layer2 = new_attention_layer(embedding_size, n_heads, 'performer')
        >>> outputs2 = layer2(inputs)
        >>> assert outputs2.shape == input_shape

    Example:
        >>> # Test with a mask
        >>> from watch.tasks.fusion.architectures.transformer import *  # NOQA
        >>> import torch
        >>> batch_size = 1
        >>> embedding_size = 4
        >>> n_heads = 2
        >>> num_tokens = 3
        >>> input_shape = (num_tokens, batch_size, embedding_size)
        >>> inputs = torch.rand(*input_shape)
        >>> key_padding_mask = torch.rand(batch_size, num_tokens) > 0.5
        >>> layer1 = new_attention_layer(embedding_size, n_heads, 'exact')
        >>> outputs1 = layer1(inputs, key_padding_mask=key_padding_mask)
    """
    if attention_impl == 'exact':
        attention = MultiheadSelfAttention(embedding_size, n_heads)
    elif attention_impl == 'performer':
        import performer_pytorch  # NOQA
        # from performer_pytorch import SelfAttention
        # attention = SelfAttention(dim=embedding_size, heads=n_heads)
        attention = FastMultiheadSelfAttention(embedding_size, n_heads)
    elif attention_impl == 'reformer':
        attention = ReformerMultiheadedSelfAttention(embedding_size, n_heads)
    else:
        raise KeyError(attention_impl)

    # num_groups = num_groups_hueristic(embedding_size)
    # norm = nn.GroupNorm(num_groups=num_groups, num_channels=embedding_size)
    norm = nn.LayerNorm(embedding_size)
    layer = ResidualAttentionSequential(
        norm,
        attention,
    )
    return layer


def new_mlp_layer(embedding_size, dropout, **kwargs):
    """
    Example:
        >>> import torch
        >>> embedding_size = 3
        >>> batch_size = 1
        >>> layer = new_mlp_layer(embedding_size, dropout=0)
        >>> input_shape = (batch_size, embedding_size)
        >>> inputs = torch.rand(*input_shape)
        >>> outputs = layer(inputs)
        >>> assert outputs.shape == (batch_size, embedding_size)
    """
    return ResidualSequential(
        nn.Linear(embedding_size, embedding_size, **kwargs),
        nn.Dropout(dropout),
        nn.GELU(),
        nn.Linear(embedding_size, embedding_size, **kwargs),
    )


class ChannelwiseTransformerEncoderLayer(nn.Module):
    """

    TODO:
        - [ ] Can we resitrict how far the spatial window looks, so it only
              sees neighboring spatial regions?

        - [ ] Flatten tokens completely and have a mask that indicates
              what tokens are allowed to see each other in each step

    Notes:
        * Currently 'mode' might indicate something like a sensor or special
          computation. Each 'mode' might have a differet number of 'features'.
          In the future this might be better specified as a dictionary that
          maps 'mode'-codes to a tensor containing only the 'features' for that
          mode. E.g.:

              inputs = {
                  'S2':        Tensor([B, T, H, W, 13]),
                  'WV':        Tensor([B, T, H, W, 8]),
                  'Latent':    Tensor([B, T, H, W, 512]),
                  'Materials': Tensor([B, T, H, W, 16]),
              }

        Currently these are all stacked into a B x T x M x H x W x max(F)
        and padded with zeros.

        Correction: the last statement is not correct.
        Curently F is hard coded to be F = 1 * ws * ws (where ws is the window
        size), so features are really spatial positions in a window. And
        the 'width' and 'height' here refer to the 'number of windows'
        in the area.

    Example:
        >>> from watch.tasks.fusion.architectures.transformer import *  # NOQA
        >>> import torch
        >>> image_size = 128
        >>> #
        >>> ws = window_size = 32
        >>> W = H = image_size // ws
        >>> B = batch_size = 2
        >>> T = num_times = 3
        >>> M = num_modes = 13  # hack for number of features in S2
        >>> F = 1 * ws * ws # hack to use spatial positions in a windows as features
        >>> input_shape = (B, T, M, H, W, F)
        >>> x = torch.rand(*input_shape)
        >>> embedding_size = F   # Embedding size must be equal to F
        >>> #
        >>> # ================================
        >>> # Joint Attentions Across all Axes
        >>> self = ChannelwiseTransformerEncoderLayer(
        >>>     axes=[('time', 'mode', 'height', 'width')],
        >>>     default_shape=['batch', 'time', 'mode', 'height', 'width', 'feature'],
        >>>     feature_axis='feature',
        >>>     batch_axis='batch',
        >>>     embedding_size=embedding_size,
        >>>     n_heads=4
        >>> )
        >>> print(self)
        >>> outputs = self(x)
        >>> assert tuple(outputs.shape) == (2, 3, 13, 4, 4, 1024)
        >>> #
        >>> # ================================
        >>> # Separable Attentions Across Time, Mode, and then Space
        >>> self = ChannelwiseTransformerEncoderLayer(
        >>>     axes=[('time', 'mode'), ('height', 'width')],
        >>>     default_shape=['batch', 'time', 'mode', 'height', 'width', 'feature'],
        >>>     feature_axis='feature',
        >>>     batch_axis='batch',
        >>>     embedding_size=embedding_size,
        >>>     n_heads=4
        >>> )
        >>> print(self)
        >>> outputs = self(x)
        >>> assert tuple(outputs.shape) == (2, 3, 13, 4, 4, 1024)
        >>> #
        >>> # ================================
        >>> # Space Only Attention
        >>> self = ChannelwiseTransformerEncoderLayer(
        >>>     axes=[('height', 'width')],
        >>>     default_shape=['batch', 'time', 'mode', 'height', 'width', 'feature'],
        >>>     feature_axis='feature',
        >>>     batch_axis='batch',
        >>>     embedding_size=embedding_size,
        >>>     n_heads=4
        >>> )
        >>> print(self)
        >>> outputs = self(x)
        >>> assert tuple(outputs.shape) == (2, 3, 13, 4, 4, 1024)
    """

    def __init__(
        self,
        axes,
        embedding_size,
        n_heads,
        dropout=0.,
        default_shape=('batch', 'time', 'mode', 'height', 'width', 'feature'),
        feature_axis='feature',
        batch_axis='batch',
        attention_impl='exact'
    ):
        super().__init__()
        self.axes = axes
        self.default_shape = default_shape
        self.feature_axis = feature_axis
        self.batch_axis = batch_axis

        self.axsep = ' '

        self.default_mask_shape = [s for s in self.default_shape
                                   if s != self.feature_axis]

        self.default_shape_str = self.axsep.join(default_shape)
        self.default_mask_shape_str = self.axsep.join(self.default_mask_shape)

        self.attention_modules = nn.ModuleDict({
            self.axsep.join(axis): new_attention_layer(
                embedding_size, n_heads, attention_impl=attention_impl)
            for axis in axes
        })
        self.mlp = new_mlp_layer(embedding_size, dropout)

    @profile
    def forward(self, inputs, flat_coordinates=None, key_padding_mask=None):
        r"""
        Args:
            x (Tensor):
                of shape B, T, M, H, W, F if flat_coordinates is unspecified
                otherwise it should be of shape N, F where N is the total
                number of tokens

            flat_coordinates (Dict[str, Tensor]):
                the time, mode, height, and width coordinate of each token
                if specified batches are unsupported

            key_padding_mask (Tensor):
                of shape B, T, M, H, W if flat_coordinates is unspecified
                otherwise should be of shape N. A True value means ignore the
                token.

        CommandLine:
            xdoctest -m watch.tasks.fusion.architectures.transformer ChannelwiseTransformerEncoderLayer.forward

        Example:
            >>> # Test that coordinate aware implementation exactly reproduces aligned variant
            >>> from watch.tasks.fusion.architectures.transformer import *  # NOQA
            >>> import numpy as np
            >>> F = embedding_size = 4
            >>> B, T, M, H, W = 1, 3, 5, 7, 11
            >>> aligned_inputs = aligned_x = (torch.rand(B, T, M, H, W, F) * 100).round() / 10
            >>> key_padding_mask = torch.rand(B, T, M, H, W) > 0.9
            >>> flat_inputs = flat_x = aligned_inputs.view(-1, embedding_size)
            >>> flat_kpm = key_padding_mask.view(-1)
            >>> #inputs = flat_inputs
            >>> flat_coordinates = None
            >>> inputs = aligned_inputs
            >>> # Test that coordinate-aware flat attention works
            >>> t_coords, m_coords, h_coords, w_coords = np.meshgrid(np.arange(T), np.arange(M), np.arange(H), np.arange(W), indexing='ij')
            >>> flat_coordinates = {
            >>>     'time':   t_coords.ravel(),
            >>>     'mode':   m_coords.ravel(),
            >>>     'height': h_coords.ravel(),
            >>>     'width':  w_coords.ravel(),
            >>> }
            >>> flat_coordinates = ub.map_vals(torch.from_numpy, flat_coordinates)
            >>> self = ChannelwiseTransformerEncoderLayer(
            >>>     #axes=[('height', 'width'), ('time',)],
            >>>     axes=[('time',)],
            >>>     default_shape=['batch', 'time', 'mode', 'height', 'width', 'feature'],
            >>>     feature_axis='feature',
            >>>     batch_axis='batch',
            >>>     embedding_size=embedding_size,
            >>>     n_heads=1
            >>> )
            >>> self = self.eval()
            >>> with torch.set_grad_enabled(False):
            >>>     print('----')
            >>>     flat_y = self.forward(flat_inputs, flat_coordinates)
            >>>     print('----')
            >>>     aligned_y = self.forward(aligned_inputs)
            >>>     print('----')
            >>>     aligned_y_mask = self.forward(aligned_inputs, key_padding_mask=key_padding_mask)
            >>>     print('----')
            >>>     flat_y_mask = self.forward(flat_inputs, flat_coordinates, key_padding_mask=flat_kpm)
            >>> print('----====-')
            >>> recon_y1 = aligned_y.view(-1, embedding_size)
            >>> recon_y1_mask = aligned_y_mask.view(-1, embedding_size)
            >>> print('flat_y=\n{!r}'.format(flat_y))
            >>> print('recon_y1=\n{!r}'.format(recon_y1))
            >>> abs_diff = (flat_y - recon_y1).abs().max()
            >>> print('abs_diff = {!r}'.format(abs_diff))
            >>> assert abs_diff < 1e-5
            >>> #
            >>> flat_y_mask.nan_to_num_()
            >>> recon_y1_mask.nan_to_num_()
            >>> abs_diff_mask = (flat_y_mask - recon_y1_mask).abs().max()
            >>> print('abs_diff_mask = {!r}'.format(abs_diff_mask))
            >>> assert abs_diff_mask < 1e-5
            >>> #flags = torch.isclose(flat_y, recon_y1)
            >>> #assert flags.all()
        """
        shape_dict = dict(zip(self.default_shape, inputs.shape))
        mask_shape_dict = ub.udict(shape_dict) - {self.feature_axis}

        if flat_coordinates:
            flat_x = inputs
        else:
            x = inputs

        previous_axial_shape = self.default_shape_str
        # previous_mask_axial_shape = self.axsep.join(self.default_shape[:-1])

        for axis in self.axes:
            if not isinstance(axis, (list, tuple)):
                axis = [axis]
            sequence_axes = self.axsep.join(axis)
            attention_layer = self.attention_modules[sequence_axes]

            if flat_coordinates is None:
                # Fast axis aligned method
                batch_axes = self.axsep.join([
                    a for a in self.default_shape
                    if (a == self.batch_axis or a not in axis) and a != self.feature_axis
                ])

                # Reshape Input to Sequence-Batch-Feature wrt to the specified axis
                # at this layer.
                axial_shape = f"({sequence_axes}) ({batch_axes}) {self.feature_axis}"
                rearrange_op = f"{previous_axial_shape} -> {axial_shape}"
                x = einops.rearrange(x, rearrange_op, **shape_dict)

                if key_padding_mask is not None:
                    mask_axial_shape = f"({batch_axes}) ({sequence_axes})"
                    mask_rearrange_op = f"{self.default_mask_shape_str} -> {mask_axial_shape}"
                    kpm = einops.rearrange(key_padding_mask, mask_rearrange_op, **mask_shape_dict)
                    # previous_mask_axial_shape = mask_axial_shape
                    y = attention_layer(x, key_padding_mask=kpm)
                else:
                    kpm = None
                    y = attention_layer(x)
                x = y
                previous_axial_shape = axial_shape
            else:
                # Batch size must be 1 in this mode.
                import numpy as np
                # New generalized coordinate method
                batch_coords = ub.dict_diff(flat_coordinates, axis)
                # axial_coords = ub.dict_subset(flat_coordinates, axis)

                # For each dimension, build groups of indices where only
                # one coordinate is varied in that dimension (e.g.
                # the same mode,x/y position over different times.
                groupids = torch.stack(list(batch_coords.values())).T.contiguous()
                groupids_numpy = groupids.numpy()
                arr = groupids_numpy
                dtype_view = np.dtype((np.void, arr.dtype.itemsize * arr.shape[1]))
                arr_view = arr.view(dtype_view)
                groupids_bytes = [r.tobytes() for r in arr_view]
                group_to_idxs = ub.group_items(range(len(groupids_bytes)), groupids_bytes)

                if key_padding_mask is None:
                    flat_kpm = None
                else:
                    flat_kpm = key_padding_mask.view(-1)

                # TODO: can we build a mask and do masked attention?
                flat_y = torch.empty_like(flat_x)
                for groupid, idxs in group_to_idxs.items():
                    x_part = flat_x[idxs]
                    x_attn = x_part[:, None, :]
                    if flat_kpm is None:
                        y_attn = attention_layer(x_attn)
                    else:
                        kpm_part = flat_kpm[idxs]
                        kpm = kpm_part[None, :]
                        y_attn = attention_layer(x_attn, key_padding_mask=kpm)
                    y_part = y_attn[:, 0, :]
                    flat_y[idxs] = y_part
                flat_x = flat_y

        if flat_coordinates is None:
            sequence_axes = self.axsep.join([
                a for a in self.default_shape
                if a not in (self.batch_axis, self.feature_axis)
            ])
            axial_shape = f"({sequence_axes}) {self.batch_axis} {self.feature_axis}"
            rearrange_op = f"{previous_axial_shape} -> {axial_shape}"
            x = einops.rearrange(x, rearrange_op, **shape_dict)
            x = self.mlp(x)
            rearrange_op = f"{axial_shape} -> {self.default_shape_str}"
            x = einops.rearrange(x, rearrange_op, **shape_dict)
            outputs = x
        else:
            flat_x = self.mlp(flat_x)
            outputs = flat_x
        return outputs


class TimmEncoder:
    """

    Example:
        >>> # xdoctest: +REQUIRES(module:timm)
        >>> from watch.tasks.fusion.architectures.transformer import *  # NOQA
        >>> import torch
        >>> in_features = 7
        >>> input_shape = B, T, M, H, W, F = (2, 3, 5, 2, 2, in_features)
        >>> inputs = torch.rand(*input_shape)
        >>> arch_name = 'vit_base_patch16_224'
        >>> self = TimmEncoder(arch_name)
    """

    def __init__(self, arch_name='vit_base_patch16_224', pretrained=True,
                 dropout=0.0, attention_impl='exact', in_features=None):
        import timm
        self.timm_model = timm.create_model(arch_name, pretrained=True)
        # embedding_size=128,
        # n_layers=4,
        # n_heads=8,
        self.timm_model
        timm.create_model('mobilenetv3_large_100_miil_in21k')
        import netharn as nh
        nh.OutputShapeFor(self.timm_model.patch_embed.proj)
        nh.OutputShapeFor(self.timm_model.blocks)
        nh.OutputShapeFor(self.timm_model.head)


class MM_VITEncoder(nn.Module):
    """
    mmsegmentation variant of VIT

    Needs 768 features.

    Notes:
        https://github.com/open-mmlab/mmsegmentation/tree/master/configs/vit

    Results:
        # 1
        https://github.com/open-mmlab/mmsegmentation/tree/master/configs/vit#ade20k
        https://github.com/open-mmlab/mmsegmentation/blob/master/configs/vit/upernet_vit-b16_mln_512x512_80k_ade20k.py
        https://github.com/open-mmlab/mmsegmentation/blob/master/configs/_base_/models/upernet_vit-b16_ln_mln.py


    Ignore:
        >>> from mmseg.models.backbones import vit
        >>> from watch.tasks.fusion.architectures.transformer import *  # NOQA
        >>> self = MM_VITEncoder()
        >>> x = torch.rand(2, 3, 768)
        >>> self.forward(x)
    """

    def __init__(self):
        super().__init__()
        from mmseg.models.backbones.vit import VisionTransformer
        kwargs = dict(
            img_size=(512, 512),
            patch_size=16,
            in_channels=3,
            embed_dims=768,
            num_layers=12,
            num_heads=12,
            mlp_ratio=4,
            out_indices=(2, 5, 8, 11),
            qkv_bias=True,
            drop_rate=0.0,
            attn_drop_rate=0.0,
            drop_path_rate=0.0,
            with_cls_token=True,
            norm_cfg=dict(type='LN', eps=1e-6),
            act_cfg=dict(type='GELU'),
            norm_eval=False,
            interpolate_mode='bicubic')
        vit_model = VisionTransformer(**kwargs)
        # We only need the encoder
        self.layers = vit_model.layers
        self.initialize_from_pretrained()
        self.in_features = self.layers[0].ln1.weight.shape[0]
        self.out_features = self.layers[-1].ffn.layers[1].out_features

    def initialize_from_pretrained(self):
        # pretrained_fpath = ub.grabdata('https://download.openmmlab.com/mmsegmentation/v0.5/vit/upernet_vit-b16_mln_512x512_80k_ade20k/upernet_vit-b16_mln_512x512_80k_ade20k_20210624_130547-0403cee1.pth')
        # FIXME: Having this import here breaks torch.package
        # not exactly sure why
        # from watch.tasks.fusion.fit import coerce_initializer
        # initializer = coerce_initializer(pretrained_fpath)
        # info = initializer.forward(self, verbose=0)  # NOQA
        ...

    def forward(self, x):
        orig_shape = x.shape
        x = x.view(x.shape[0], -1, x.shape[1])
        for i, layer in enumerate(self.layers):
            x = layer(x)
            # if i == len(self.layers) - 1:
            #     if self.final_norm:
            #         x = self.norm1(x)
            # if i in self.out_indices:
            #     if self.with_cls_token:
            #         # Remove class token and reshape token for decoder head
            #         out = x[:, 1:]
            #     else:
            #         out = x
            #     B, _, C = out.shape
            #     out = out.reshape(B, hw_shape[0], hw_shape[1],
            #                       C).permute(0, 3, 1, 2).contiguous()
            #     if self.output_cls_token:
            #         out = [out, x[:, 0]]
            #     outs.append(out)
        x = x.view(*orig_shape[0], x.shape[-1])
        return x


class DeiTEncoder(nn.Module):
    """
    https://github.com/rishikksh20/ViViT-pytorch
    https://pytorch.org/tutorials/beginner/vt_tutorial.html

    Example:
        >>> # xdoctest: +REQUIRES(module:timm)
        >>> from watch.tasks.fusion.architectures.transformer import *  # NOQA
        >>> import torch
        >>> in_features = 7
        >>> input_shape = B, T, M, H, W, F = (2, 3, 5, 2, 2, in_features)
        >>> inputs = torch.rand(*input_shape)
        >>> self = DeiTEncoder(in_features)
        >>> outputs = self.forward(inputs)
    """

    def __init__(self, in_features):
        super().__init__()
        deit = torch.hub.load('facebookresearch/deit:main', 'deit_base_patch16_224', pretrained=True)
        blocks = deit.blocks
        block_in_features = blocks[0].norm1.weight.shape[0]
        self.first = nn.Linear(in_features, out_features=block_in_features)
        self.blocks = blocks
        self.in_features = in_features
        self.out_features = blocks[-1].mlp.fc2.out_features

    def forward(self, inputs):
        B, T, M, H, W, F = inputs.shape
        x = einops.rearrange(inputs, 'b t m h w f -> b (t m h w) f')
        x = self.first(x)
        x = self.blocks(x)
        outputs = einops.rearrange(x, 'b (t m h w) f -> b t m h w f', t=T, m=M, h=H, w=W)
        return outputs


class PerceiverEncoder(nn.Module):
    """
    https://github.com/lucidrains/perceiver-pytorch/blob/main/perceiver_pytorch/perceiver_io.py

    Example:
        >>> from watch.tasks.fusion.architectures.transformer import PerceiverEncoder  # NOQA
        >>> import torch
        >>> B, T, M, H, W, F = 1, 2, 3, 5, 8, 13
        >>> self = PerceiverEncoder(F, dropout=0.1)
        >>> inputs = torch.rand(B, T, M, H, W, F)
        >>> outputs = self(inputs)
        >>> assert outputs.shape == (B, T, M, H, W, F)
    """

    def __init__(self, in_features, depth=4, dropout=0.0):
        super().__init__()
        import perceiver_pytorch as perceiver
        # No dropout in perceiver? Perform it on input tokens.
        self.dropout = nn.Dropout(dropout)
        self.perceiver = perceiver.PerceiverIO(
            depth=depth,
            dim=in_features,
            queries_dim=in_features,
            num_latents=512,
            latent_dim=256,
            cross_heads=1,
            latent_heads=8,
            cross_dim_head=64,
            latent_dim_head=64,
            weight_tie_layers=False,
            decoder_ff=False,
            logits_dim=None,
        )
        self.in_features = in_features
        self.out_features = in_features

    def forward(self, inputs):
        B, T, M, H, W, F = inputs.shape
        x = einops.rearrange(inputs, 'b t m h w f -> b (t m h w) f')
        x = self.dropout(x)
        x = self.perceiver(x, queries=x)
        outputs = einops.rearrange(x, 'b (t m h w) f -> b t m h w f', t=T, m=M, h=H, w=W)
        return outputs


class FusionEncoder(nn.Module):
    """
    Primary entry point to create a feature transformer

    Performs multiple "channelwise" (maybe rename to axil?) attention
    encodings in a row

    Example:
        >>> from watch.tasks.fusion.architectures.transformer import *  # NOQA
        >>> import torch
        >>> in_features = 7
        >>> input_shape = B, T, M, H, W, F = (2, 3, 5, 2, 2, in_features)
        >>> inputs = torch.rand(*input_shape)
        >>> model = FusionEncoder(
        >>>     in_features=in_features,
        >>>     axes=[("time", "mode", "height", "width")],
        >>>     default_shape=["batch", "time", "mode", "height", "width", "feature"],
        >>>     feature_axis="feature",
        >>>     batch_axis="batch",
        >>>     n_layers=8,
        >>>     embedding_size=256,
        >>>     n_heads=4
        >>> )
        >>> model(inputs)
        >>> output = model(inputs)
        >>> assert output.shape == (2, 3, 5, 2, 2, 256)
        >>> #
        >>> # Test Lazy variant
        >>> model = FusionEncoder(
        >>>     in_features=None,
        >>>     axes=[("time", "mode", "height", "width")],
        >>>     default_shape=["batch", "time", "mode", "height", "width", "feature"],
        >>>     feature_axis="feature",
        >>>     batch_axis="batch",
        >>>     n_layers=8,
        >>>     embedding_size=256,
        >>>     n_heads=4
        >>> )
        >>> print(model)
        >>> inputs = torch.rand(*input_shape)
        >>> output = model(inputs)
        >>> assert output.shape == (2, 3, 5, 2, 2, 256)

    Ignore:
        traced = torch.jit.trace(model, inputs)
        import timerit
        ti = timerit.Timerit(5, bestof=1, verbose=2)
        for timer in ti.reset('time'):
            model(inputs)
        for timer in ti.reset('time'):
            traced(inputs)

    Ignore:
        >>> # Get a sense of the arch size
        >>> from watch.tasks.fusion.architectures.transformer import *  # NOQA
        >>> rows = []
        >>> import netharn as nh  # NOQA
        >>> for key, config in ub.ProgIter(list(encoder_configs.items())):
        >>>     self = FusionEncoder(in_features=256, **config)
        >>>     num_params = nh.util.number_of_parameters(self)
        >>>     row = {'arch': key, 'num_params': num_params}
        >>>     row.update(config)
        >>>     print('row = {}'.format(ub.urepr(row, nl=0, sort=0)))
        >>>     rows.append(row)
        >>> import pandas as pd
        >>> data = pd.DataFrame(rows).sort_values('num_params')
        >>> print(data.to_string())

        >>> # Look at only smt configs
        >>> flags = data['axes'].apply(lambda x: x == [("height", "width"), ("time",), ("mode",)])
        >>> print(data[flags].to_string())
    """

    def __init__(self, axes,
                 default_shape=('batch', 'sequence', 'feature'),
                 feature_axis='feature',
                 batch_axis='batch',
                 embedding_size=128,
                 n_layers=4,
                 n_heads=8,
                 dropout=0.0,
                 attention_impl='exact',
                 in_features=None):
        super().__init__()
        if in_features is None:
            # Use lazy linear to allow data to specify the channel dims
            first = nn.LazyLinear(embedding_size)
        else:
            first = nn.Linear(in_features=in_features, out_features=embedding_size)

        _layers = [
            ChannelwiseTransformerEncoderLayer(
                axes,
                embedding_size=embedding_size,
                n_heads=n_heads,
                dropout=dropout,
                default_shape=default_shape,
                feature_axis=feature_axis,
                batch_axis=batch_axis,
                attention_impl=attention_impl,
            )
            for _ in range(n_layers)
        ]

        self.in_features = in_features
        self.out_features = embedding_size

        self.first = first
        self.layers = nn.Sequential(*_layers)

    def forward(self, x, flat_coordinates=None, key_padding_mask=None, mask=None):
        if mask is not None:
            key_padding_mask = mask
        x = self.first(x)
        for layer in self.layers:
            # Can't use sequentail because we need extra args
            x = layer(x, flat_coordinates=flat_coordinates,
                      key_padding_mask=key_padding_mask)
        return x


def _build_global_configs():
    """
    Previously we manually defined a bunch of functions, now it
    is defined programatically
    """
    # dont define tons of functions, use a configuration dictionary
    _smt_axes_basis = dict(
        joint=[('time', 'mode', 'height', 'width')],
        stm=[('height', 'width'), ('time',), ('mode',)],
        sm=[('height', 'width'), ('mode',)],
        st=[('height', 'width'), ('time',)],
        tm=[('time',), ('mode',)],
        s=[('height', 'width')],
        t=[('time',)],
        hwtm=[('height',), ('width',), ('time',), ('mode',)],
        m=[('mode',)],
    )

    _encoder_size_basis = {
        'p1': dict(n_layers=1, embedding_size=64, n_heads=4),
        'p2': dict(n_layers=2, embedding_size=64, n_heads=4),
        'p2w': dict(n_layers=2, embedding_size=128, n_heads=8),

        'p3': dict(n_layers=3, embedding_size=128, n_heads=4),
        'p4': dict(n_layers=4, embedding_size=128, n_heads=4),

        'p8': dict(n_layers=8, embedding_size=128, n_heads=4),
        'p16': dict(n_layers=16, embedding_size=128, n_heads=4),

        'b8': dict(n_layers=8, embedding_size=384, n_heads=4),
        'n12': dict(n_layers=12, embedding_size=128, n_heads=4),
        't12': dict(n_layers=12, embedding_size=192, n_heads=4),
        't24': dict(n_layers=24, embedding_size=192, n_heads=4),
        's12': dict(n_layers=12, embedding_size=384, n_heads=8),
        's24': dict(n_layers=24, embedding_size=384, n_heads=8),
        'm24': dict(n_layers=24, embedding_size=512, n_heads=8),
        'l24': dict(n_layers=24, embedding_size=768, n_heads=8),
    }

    # space-mode-time transformer params
    _smt_value = dict(
        default_shape=['batch', 'time', 'mode', 'height', 'width', 'feature'],
        feature_axis='feature',
        batch_axis='batch',
    )

    encoder_configs = {}
    for axes_code, axes_value in _smt_axes_basis.items():
        for size_code, size_value in _encoder_size_basis.items():
            code = f'smt_it_{axes_code}_{size_code}'
            encoder_configs[code] = ub.dict_union(
                size_value, _smt_value, dict(axes=axes_value))

    # space-mode transformer params
    _sm_value = dict(
        default_shape=['batch', 'mode', 'height', 'width', 'feature'],
        feature_axis='feature',
        batch_axis='batch',
    )

    _sm_axes_basis = {
        'joint': [('mode', 'height', 'width')],
        'sm': [('height', 'width'), ('mode',)],
    }

    for axes_code, axes_value in _sm_axes_basis.items():
        for size_code, size_value in _encoder_size_basis.items():
            code = f'sm_it_{axes_code}_{size_code}'
            encoder_configs[code] = ub.dict_union(
                size_value, _sm_value, dict(axes=axes_value))
    return encoder_configs


encoder_configs = _build_global_configs()

# print('encoder_configs = {}'.format(ub.urepr(list(encoder_configs.keys(), nl=1)))

# ========================================
# Below is an implementation of the transformer architecture that uses the
# same base components and follows the same patterns as the perceiver. This
# is a 1-to-1 drop-in replacement for perceiver models (minus latent_dim/num_latents options).
# ========================================


def default(val, d):
    return val if val is not None else d


def cache_fn(f):
    cache = None

    @wraps(f)
    def cached_fn(*args, _cache=True, **kwargs):
        if not _cache:
            return f(*args, **kwargs)
        nonlocal cache
        if cache is not None:
            return cache
        cache = f(*args, **kwargs)
        return cache
    return cached_fn

# helper classes


class PreNorm(nn.Module):
    def __init__(self, dim, fn, context_dim=None):
        super().__init__()
        self.fn = fn
        self.norm = nn.LayerNorm(dim)
        self.norm_context = nn.LayerNorm(context_dim) if context_dim is not None else None

    def forward(self, x, **kwargs):
        x = self.norm(x)

        if self.norm_context is not None:
            context = kwargs['context']
            normed_context = self.norm_context(context)
            kwargs.update(context=normed_context)

        return self.fn(x, **kwargs)


class GEGLU(nn.Module):
    def forward(self, x):
        x, gates = x.chunk(2, dim=-1)
        return x * F.gelu(gates)


class FeedForward(nn.Module):
    def __init__(self, dim, mult=4):
        super().__init__()
        self.net = nn.Sequential(
            nn.Linear(dim, dim * mult * 2),
            GEGLU(),
            nn.Linear(dim * mult, dim)
        )

    def forward(self, x):
        return self.net(x)


class Attention(nn.Module):
    def __init__(self, query_dim, context_dim=None, output_dim=None, heads=8, dim_head=64):
        super().__init__()
        inner_dim = dim_head * heads
        context_dim = default(context_dim, query_dim)
        output_dim = default(output_dim, query_dim)
        self.scale = dim_head ** -0.5
        self.heads = heads

        self.to_q = nn.Linear(query_dim, inner_dim, bias=False)
        self.to_kv = nn.Linear(context_dim, inner_dim * 2, bias=False)
        self.to_out = nn.Linear(inner_dim, output_dim)

    def forward(self, x, context=None, mask=None):
        h = self.heads

        q = self.to_q(x)
        context = default(context, x)
        k, v = self.to_kv(context).chunk(2, dim=-1)

        q, k, v = map(lambda t: rearrange(t, 'b n (h d) -> (b h) n d', h=h), (q, k, v))

        sim = einsum('b i d, b j d -> b i j', q, k) * self.scale

        if mask is not None:
            mask = rearrange(mask, 'b ... -> b (...)')
            max_neg_value = -torch.finfo(sim.dtype).max
            mask = repeat(mask, 'b j -> (b h) () j', h=h)
            sim.masked_fill_(~mask, max_neg_value)

        # attention, what we cannot get enough of
        attn = sim.softmax(dim=-1)

        out = einsum('b i j, b j d -> b i d', attn, v)
        out = rearrange(out, '(b h) n d -> b n (h d)', h=h)
        return self.to_out(out)


class BackboneEncoderDecoder:
    pass


class TransformerEncoderDecoder(nn.Module, BackboneEncoderDecoder):
    def __init__(
        self,
        encoder_depth: int = 2,
        decoder_depth: int = 1,
        dim: int = 128,
        queries_dim: int = 96,
        logits_dim: int = 32,
        decode_cross_every: int = 1,
        cross_heads: int = 1,
        latent_heads: int = 8,
        cross_dim_head: int = 64,
        latent_dim_head: int = 64,
        weight_tie_layers: bool = False,
    ):
        super().__init__()

        self.has_decoder = (decoder_depth > 0) and (queries_dim is not None) and (queries_dim > 0)

        def latent_cross_attn():
            return PreNorm(
                queries_dim,
                Attention(
                    queries_dim, dim, dim,
                    heads=cross_heads,
                    dim_head=cross_dim_head),
                context_dim=dim,
            )

        def latent_attn():
            return PreNorm(
                dim,
                Attention(
                    dim,
                    heads=latent_heads,
                    dim_head=latent_dim_head),
            )

        def latent_ff():
            return PreNorm(
                dim,
                FeedForward(dim),
            )

        cache_args = {'_cache': weight_tie_layers}

        get_latent_attn, get_latent_ff = map(cache_fn, (latent_attn, latent_ff))
        self.encoder_layers = nn.ModuleList([])
        for i in range(encoder_depth):
            self.encoder_layers.append(nn.ModuleList([
                get_latent_attn(**cache_args),
                get_latent_ff(**cache_args)
            ]))

        if self.has_decoder:
            get_latent_cross_attn, get_latent_attn, get_latent_ff = map(cache_fn, (latent_cross_attn, latent_attn, latent_ff))
            self.decoder_layers = nn.ModuleList([])
            for i in range(decoder_depth):
                if (i % decode_cross_every) == 0:
                    layers = [
                        get_latent_cross_attn(**cache_args),
                        get_latent_attn(**cache_args),
                        get_latent_ff(**cache_args)
                    ]
                else:
                    layers = [
                        get_latent_attn(**cache_args),
                        get_latent_ff(**cache_args)
                    ]
                self.decoder_layers.append(nn.ModuleList(layers))

        self.to_logits = nn.Linear(dim, logits_dim) if logits_dim is not None else nn.Identity()

    def forward(
        self,
        x,
        mask=None,
        queries=None
    ):
        b = x.shape[0]

        # layers
        for self_attn, self_ff in self.encoder_layers:
            x = self_attn(x, mask=mask) + x
            x = self_ff(x) + x

        if queries is None or not self.has_decoder:
            return self.to_logits(x)

        # make sure queries contains batch dimension
        if queries.ndim == 2:
            queries = repeat(queries, 'n d -> b n d', b=b)

        # cross attend from decoder queries to latents
        for layers in self.decoder_layers:
            if len(layers) == 3:
                cross_attn, self_attn, self_ff = layers
                x = cross_attn(queries, context=x)
            else:
                self_attn, self_ff = layers

            x = self_attn(x) + x
            x = self_ff(x) + x

        # final linear out
        return self.to_logits(x)


class MM_VITEncoderDecoder(nn.Module, BackboneEncoderDecoder):
    """
    mmsegmentation variant of VIT

    Needs 768 features.

    Notes:
        https://github.com/open-mmlab/mmsegmentation/tree/master/configs/vit

    Results:
        # 1
        https://github.com/open-mmlab/mmsegmentation/tree/master/configs/vit#ade20k
        https://github.com/open-mmlab/mmsegmentation/blob/master/configs/vit/upernet_vit-b16_mln_512x512_80k_ade20k.py
        https://github.com/open-mmlab/mmsegmentation/blob/master/configs/_base_/models/upernet_vit-b16_ln_mln.py


    Example:
        >>> # xdoctest: +REQUIRES(module:mmseg)
        >>> from watch.tasks.fusion.architectures.transformer import *  # NOQA
        >>> self = MM_VITEncoderDecoder(16, 16, 16)
        >>> x = torch.rand(2, 3, 16)
        >>> self.forward(x)

    Ignore:
        >>> # xdoctest: +REQUIRES(module:mmseg)
        >>> # This tests downloading weights from the MM repo
        >>> from watch.tasks.fusion.architectures.transformer import *  # NOQA
        >>> self = MM_VITEncoderDecoder(16, 16, 16, pretrained="upernet_vit-b16_mln_512x512_80k_ade20k")
        >>> x = torch.rand(2, 3, 16)
        >>> self.forward(x)
    """

    pretrained_fpath_shortnames = {
        "upernet_vit-b16_mln_512x512_80k_ade20k":
            'https://download.openmmlab.com/mmsegmentation/v0.5/vit/upernet_vit-b16_mln_512x512_80k_ade20k/upernet_vit-b16_mln_512x512_80k_ade20k_20210624_130547-0403cee1.pth',
    }

    def __init__(
        self,
        dim,
        logits_dim,
        queries_dim=None,
        pretrained=None,
    ):
        from mmseg.models.backbones.vit import VisionTransformer
        super().__init__()

        # if a short name is used, replace it with the appropriate full path
        if pretrained in MM_VITEncoderDecoder.pretrained_fpath_shortnames.keys():
            pretrained = MM_VITEncoderDecoder.pretrained_fpath_shortnames[pretrained]
            pretrained = ub.grabdata(pretrained)

        kwargs = dict(
            pretrained=pretrained,
            img_size=(512, 512),
            patch_size=16,
            in_channels=3,
            embed_dims=768,
            num_layers=12,
            num_heads=12,
            mlp_ratio=4,
            out_indices=(2, 5, 8, 11),
            qkv_bias=True,
            drop_rate=0.0,
            attn_drop_rate=0.0,
            drop_path_rate=0.0,
            with_cls_token=True,
            norm_cfg=dict(type='LN', eps=1e-6),
            act_cfg=dict(type='GELU'),
            norm_eval=False,
            interpolate_mode='bicubic')
        vit_model = VisionTransformer(**kwargs)
        # We only need the encoder
        self.layers = vit_model.layers

        # if a pretrained path is provided, try to use it
        # if isinstance(pretrained, str):
        #     self.initialize_from_pretrained(pretrained)

        self.encoder_in_features = self.layers[0].ln1.weight.shape[0]
        self.encoder_out_features = self.layers[-1].ffn.layers[1].out_features

        self.input_projector = nn.Linear(dim, self.encoder_in_features)
        self.output_projector = nn.Linear(self.encoder_out_features, logits_dim)

        self.has_decoder = (queries_dim is not None) and (queries_dim > 0)

        if self.has_decoder:
            self.query_projector = nn.Linear(queries_dim, self.encoder_out_features)
            self.decoder = nn.TransformerDecoder(
                nn.TransformerDecoderLayer(d_model=self.encoder_out_features, nhead=8, dim_feedforward=512, batch_first=True),
                num_layers=1,
            )

    def initialize_from_pretrained(self, fpath):
        # FIXME: Having this import here breaks torch.package
        # initializer = coerce_initializer(fpath)
        # info = initializer.forward(self, verbose=0)  # NOQA
        pass

    def forward(self, x, mask=None, queries=None):
        # orig_shape = x.shape
        # x = x.view(x.shape[0], -1, x.shape[1])
        x = self.input_projector(x)
        for i, layer in enumerate(self.layers):
            x = layer(x)
            # if i == len(self.layers) - 1:
            #     if self.final_norm:
            #         x = self.norm1(x)
            # if i in self.out_indices:
            #     if self.with_cls_token:
            #         # Remove class token and reshape token for decoder head
            #         out = x[:, 1:]
            #     else:
            #         out = x
            #     B, _, C = out.shape
            #     out = out.reshape(B, hw_shape[0], hw_shape[1],
            #                       C).permute(0, 3, 1, 2).contiguous()
            #     if self.output_cls_token:
            #         out = [out, x[:, 0]]
            #     outs.append(out)
        # x = x.view(*orig_shape[0], x.shape[-1])

        if (queries is None) or (not self.has_decoder):
            return x

        queries = self.query_projector(queries)
        x = self.decoder(queries, x)
        x = self.output_projector(x)

        return x
